import numpy as np
import pyarrow as pa

import os

from setuptools.extension import Extension
from Cython.Build import cythonize

def build(setup_kwargs):
    """
    This is a callback for poetry used to hook in our extensions.
    """
    ext_modules = cythonize("src/phantombuster/merge_cython.pyx")

    import pyarrow
    pyarrow.create_library_symlinks()

    for ext in ext_modules:
        # The Numpy C headers are currently required
        ext.include_dirs.append(np.get_include())
        ext.include_dirs.append(pa.get_include())
        ext.libraries.extend(pa.get_libraries())
        ext.library_dirs.extend(pa.get_library_dirs())

        if os.name == 'posix':
            ext.extra_compile_args.append('-std=c++17')
        ext.runtime_library_dirs.append("$ORIGIN/../pyarrow")

        # Try uncommenting the following line on Linux
        # if you get weird linker errors or runtime crashes
        ext.define_macros.append(("_GLIBCXX_USE_CXX11_ABI", "0"))


    setup_kwargs.update(
        {
            # declare the extension so that setuptools will compile it
            "ext_modules": ext_modules,
        }
    )



import zlib
import re
import numpy as np


class EndOfFile(Exception):
    pass

def read_header(f):
    header = f.read(12)
    if header == b"":
        raise EndOfFile()
    xlen = parse_xlen(header)

    subfields = f.read(xlen)
    if subfields == b"":
        raise EndOfFile()

    bsize = parse_bsize(subfields, xlen)
    return xlen, bsize

def read_block(f, bsize, xlen):
    """Assumes you called read_header before"""
    rest = f.read(bsize - 12 - xlen)
    return zlib.decompress(rest, -15)

def parse_xlen(header):
    return int.from_bytes(header[-2:], "little")

def parse_bsize(subfields, xlen):
    offset = 0
    block_size = None
    while offset <= xlen:
        if subfields[offset:offset+2] == b"BC":
            block_size = int.from_bytes(subfields[offset+4:offset+6], "little") +1
            break
        else:
            offset += int.from_bytes(subfields[offset+2:offset+4], "little") + 4
    if block_size is None:
        raise Exception("Could not find BSIZE")

    return block_size

def parse_read_start(data):
    if data == b"":
        raise EndOfFile()

    found = None
    i = 0
    for m in re.finditer(8*b"\xff", data):
        if found:
            offset = m.start() - found 
            if offset == 20 and (found-4 >= 0):
                return found-4 
            else:
                found = m.start()
        else:
            found = m.start()
        i += 1
        if i == 10:
            raise Exception("Could not find the beginning of the read, either read contains FF bytes or assumptions are violated")
    raise Exception("Could not find the beginning of the read, assumptions are violated")

def index(f, every):
    coffset = 0
    index = []
    global_i = 0
    i = 0
    first_block = True

    while True:
        i += 1
        global_i += 1
        f.seek(coffset)
        try:
            xlen, bsize = read_header(f)
        except EndOfFile:
            break

        #if True:
        if i % every == 0 or global_i <= 2:
            block_data = read_block(f, bsize, xlen)
            if first_block:
                uoffset = 0
            else:
                try:
                    uoffset = parse_read_start(block_data)
                except EndOfFile:
                    break
            idx = (coffset << 16) | uoffset
            if idx >= 0:
                index.append(idx)

        coffset += bsize
        first_block = False

    index = index[1:] + [np.inf]

    return index





import regex
import json
import phantombuster.plumbing as plumbing
import phantombuster.config_files

from .plumbing import pairwise
from phantombuster import bamindexer
from phantombuster import statistics
from phantombuster.io_ import write_parquet

import os
import logging
from collections import defaultdict, namedtuple, Counter
import gzip

import pyarrow as pa
import pyarrow.types
import pyarrow.csv
import pyarrow.parquet
import numpy as np

import polars as pl


class PhantomBusterException(Exception):
    pass

class PhantomBusterUnknownFileType(PhantomBusterException):
    pass


def index_group(group, every):
    if len(group.files) == 1:
        return index_file(group.files[0].path, every)
    else:
        idxs = [[0]* len(group.files)] + [[np.inf]* len(group.files)]
        return list(zip(*idxs))


def index_file(file, every):
    if file.endswith('.bam'):
        return index_bamfile(file, every)
    elif file.endswith('.fastq.gz'):
        return index_fastqgz_file(file, every)

def index_bamfile(bamfile, every):
    with open(bamfile, mode="rb") as f:
        return bamindexer.index(f, every)


def _read_complete_read(f, read_size=1024):
    buffer = []
    lines = 0
    tries = 0
    while lines < 7 and tries < 100:
        read = f.read(read_size)
        if len(read) == 0:
            raise EOFError()
        lines += read.count(b'\n')
        buffer.append(read)
        tries += 1
    if tries >= 100:
        raise ValueError("Can't find reads in fastq file")
    complete_read = b''.join(buffer)
    lines = complete_read.split(b'\n')
    for line_idx in range(5):
        if len(lines[line_idx]) > 0 and lines[line_idx][0] == b'@'[0]:
            break
        else:
            pass
            #print('no @ in', lines[line_idx][0])
    else:
        raise EOFError()

    single_read = b'\n'.join(lines[line_idx:line_idx+4])
    offset = complete_read.find(single_read)
    return single_read, offset


def index_fastqgz_file(file, every):
    with gzip.open(file) as f:
        read, offset = _read_complete_read(f)
        total_readsize = len(read)
        reads_analysed = 1

        readsize = int(total_readsize / reads_analysed)

        idx = 0
        indices = []
        while True:
            try:
                f.seek(idx)
                read, offset = _read_complete_read(f, readsize)

                indices.append(idx+offset)
                idx = idx + offset + every

                total_readsize += len(read)
                reads_analysed += 1
                readsize = int(total_readsize / reads_analysed)

            except EOFError:
                break
        indices.append(np.inf)
    return indices


def demultiplex(input_files_file, outpath, outpath_stats, regex_file, barcode_hierarchy_file, debug, show_qc, scheduler):

    logger = logging.getLogger("phantombuster")

    input_groups = phantombuster.config_files.read_input_files_file(input_files_file)

    logger.info(f"Deduplicating {len(input_groups)} many groups")
    logger.info("Reading mappings")

    # read mapping files
    barcode_hierarchy = phantombuster.config_files.read_barcode_hierarchy_file(barcode_hierarchy_file)

    regex_dictionary = phantombuster.config_files.read_regex_file(regex_file)

    logger.info("Scheduling indexing %s group(s)", len(input_groups))
    # BGZF blocks per task, each block is <64KB so 20000 should roughly be 1.28 GB
    if scheduler.query_workers() == 0:
        logger.warning("Work is getting scheduled and no workers are currently connected. Launch workers now.")

    idxs_promises = []
    for group in input_groups:
        promise = scheduler.schedule(index_group, group, 20000)
        idxs_promises.append(promise)

    for promise in idxs_promises:
        r = 0
        waiting_time = 0
        waiting_notice_time = 10
        while isinstance(r, int):
            waiting_time += r
            if waiting_time > waiting_notice_time:
                waiting_time = 0
                number_of_workers = scheduler.query_workers()
                if number_of_workers == 0:
                    logger.warning("Currently no workes are connected to the server. Server will make no progress without workers.")
                    if waiting_notice_time == 10:
                        waiting_notice_time = 60
                    elif waiting_notice_time == 60:
                        waiting_notice_time = 10*60
                else:
                    waiting_notice_time = 10*60
            r = scheduler.wait(promise, return_after_wait=True)

    idxs = [scheduler.wait(promise, unpack=True) for promise in idxs_promises]

    logger.info("Parsing file(s) piecewise, %s many subtasks", sum(len(idx) for idx in idxs))
    logger.debug("Index is %s", idxs)
    outs = []
    for group, index in zip(input_groups, idxs):
        for section in pairwise(index):
            logger.debug("Scheduling deduplicate section %s", section)
            t = scheduler.schedule(plumbing.deduplicate_section, group, section, regex_dictionary, barcode_hierarchy, show_qc)
            outs.append(t)
    scheduler.wait_all(outs, unpack=False)

    logger.info("Extracted reads from individual files, combining %s many files", len(outs))
    new_outs = []
    while len(outs) > 1:
        while len(outs) > 1:
            a = outs.pop(0)
            b = outs.pop(0)
            logger.debug("Waiting on two files")
            scheduler.wait(a, unpack=False)
            scheduler.wait(b, unpack=False)
            logger.debug("Scheduling merging of two files, %s remaining.", len(outs))
            result = scheduler.schedule(plumbing.combine, [a, b], barcode_hierarchy)
            new_outs.append(result)
        if len(outs) == 1:
            logger.debug("Cannot merge, only one file remaining")
            new_outs.append(outs[0])
        logger.debug("One round done, there were %s many files, now there are %s", len(outs), len(new_outs))
        outs = new_outs
        new_outs = []

    logger.debug("Waiting for last combination round")
    out = scheduler.wait(outs[-1], unpack=True)
    if out is None or out[0] is None:
        logger.error("Not a single read was extracted. This might indicates a problem with the references or a missconfiguration of the regexes. Aborting. ")
        logger.debug('Result of combining all files is: %s', out)
        raise Exception("Not a single read extracted")

    write_parquet(out[0], outpath)

    with open(outpath_stats, mode="w") as f:
        json.dump(out[1], f)

    logger.info("Demultiplexing done")
    return out

def read_file(path):
    if path.endswith(".csv"):
        df = pa.csv.read_csv(path)
    elif path.endswith(".parquet"):
        df = pa.parquet.read_table(path)
    else:
        raise PhantomBusterUnknownFileType("File {threshold_file} has unknown file type. Supported are .csv and .parquet.")
    return df

def error_correct(samplefile, outfilename, threshold, barcode_hierarchy, project, remove_ambigious):

    logger = logging.getLogger('phantombuster')

    logger.debug(f"Reading table {samplefile}")
    df = pl.scan_parquet(samplefile)
    number_of_reads, number_of_lineages = df.select(pl.col('reads').sum(), pl.len()).collect()
    number_of_reads, number_of_lineages = number_of_reads[0], number_of_lineages[0]
    logger.debug(f"Table has {number_of_lineages} many rows and {number_of_reads} many reads")

    first_bc = barcode_hierarchy[0]

    if first_bc["type"] == 'reference':
        logger.debug(f"First barcode in hierarchy '{barcode_hierarchy[0]['name']}' is of 'reference' type, partitioning")
        partitions = df.select(pl.col(first_bc['name']).unique().alias('values')).collect()['values'].to_list()
        logger.debug(f"Partitioning Done, created {len(partitions)} many partitions.")
    else:
        logger.debug(f"First barcode in hierarchy '{barcode_hierarchy[0]['name']}' is of 'random' type, no partitioning performed")
        partitions = [None]

    scheduler = project.get_scheduler()
    with scheduler as scheduler:
        logger.debug('Distributing each partition to one worker')
        tasks = []
        for value in partitions:
            task = scheduler.schedule(error_correct_partition, samplefile, value, None, threshold, barcode_hierarchy, use_column=first_bc['name'], remove_ambigious=remove_ambigious)
            tasks.append(task)
        logger.debug('Distributed all partitions, waiting for workers')
        results = scheduler.gather(*tasks)
        logger.debug('Workers done, performing postprocessing')

    tables = [table for table, stats in results]

    cast_tables = []
    for table in tables:
        for column_name in table.column_names:
            if pa.types.is_string(table[column_name].type):
                table = table.set_column(table.column_names.index(column_name), column_name, table[column_name].cast("large_string"))
            if pa.types.is_binary(table[column_name].type):
                table = table.set_column(table.column_names.index(column_name), column_name, table[column_name].cast("large_binary"))
        cast_tables.append(table)
    tables = cast_tables

    corrected_table = pyarrow.concat_tables(tables)

    number_of_lineages_corrected = len(corrected_table)
    number_of_reads_corrected = pa.compute.sum(corrected_table["reads"]).as_py()

    write_parquet(corrected_table, outfilename)

    return {"sample": samplefile,
            "lids_uncorrected": number_of_lineages,
            "lids_corrected": number_of_lineages_corrected,
            "reads_uncorrected": number_of_reads,
            "reads_corrected": number_of_reads_corrected}

def error_correct_partition(samplefile, lower_or_value, length_opt, threshold, barcode_hierarchy, use_column=False, remove_ambigious=True):

    logger = logging.getLogger("phantombuster")

    if lower_or_value is None:
        logger.info(f'Read table from {samplefile} completly')
        table = pl.scan_parquet(samplefile).collect().to_arrow()
    elif use_column:
        assert isinstance(use_column, str)
        assert isinstance(lower_or_value, str)
        value = lower_or_value
        table = pl.scan_parquet(samplefile).filter(pl.col(use_column) == value).collect().to_arrow()
        logger.info(f'Read table from {samplefile} with {use_column} only {value}.')
    else:
        assert isinstance(lower_or_value, int)
        lower = lower_or_value
        assert length_opt is not None
        length = length_opt

        table = pa.parquet.read_table(samplefile, memory_map=True, use_threads=False)
        table = table.slice(lower, length)
        logger.info(f'Read table from {samplefile} from {lower} with length {length}.')

    number_of_lineages = len(table)
    number_of_reads = pa.compute.sum(table["reads"]).as_py()

    logger.info('Starting error correction')
    corrected_table = plumbing.error_correct(table, barcode_hierarchy, threshold)
    logger.info('Done error correction')

    if remove_ambigious:
        logger.info('Removing barcode sequences with ambigious characters (NRYSWKMBDHV).')
        corrected_table = plumbing.remove_ambigious(corrected_table, barcode_hierarchy)

    number_of_lineages_corrected = len(corrected_table)
    number_of_reads_corrected = pa.compute.sum(corrected_table["reads"]).as_py()

    stats = {"lids_uncorrected": number_of_lineages,
            "lids_corrected": number_of_lineages_corrected,
            "reads_uncorrected": number_of_reads,
            "reads_corrected": number_of_reads_corrected}

    return corrected_table, stats

def _read_mappingfile(mappingfile):
    table = pa.csv.read_csv(mappingfile)
    map = dict(zip(table["sample"].to_pylist(), table["group"].to_pylist()))
    return map

def hopping_removal(input_file, alpha_threshold, hopping_barcodes):
    logger = logging.getLogger("phantombuster")
    master_table = pa.parquet.read_table(input_file)

    number_of_lineages = len(master_table)
    number_of_reads = pa.compute.sum(master_table["reads"]).as_py()

    for column_name in master_table.column_names:
        if pa.types.is_string(master_table[column_name].type):
            logger.warning(f"Detected datatype 'string' in column {column_name}, casting to 'large_string'")
            master_table = master_table.set_column(master_table.column_names.index(column_name), column_name, master_table[column_name].cast("large_string"))
        if pa.types.is_binary(master_table[column_name].type):
            logger.warning(f"Detected datatype 'binary' in column {column_name}, casting to 'large_binary'")
            master_table = master_table.set_column(master_table.column_names.index(column_name), column_name, master_table[column_name].cast("large_binary"))

    logger.info("Starting Hopping Removal")
    threshold_list = []
    for barcode in hopping_barcodes:
        logger.info(f"Considering Barcode '{barcode}'")
        threshold = plumbing.calculate_hopping_threshold(master_table, barcode, alpha_threshold=alpha_threshold)
        threshold_list.append(threshold)
    master_table = plumbing.call_hopping(master_table, threshold_list)

    logger.info("Hopping Removal Done")

    number_of_lineages_corrected = len(master_table)
    number_of_reads_corrected = pa.compute.sum(master_table["reads"]).as_py()

    stats = {"lids_before": number_of_lineages,
             "lids_after": number_of_lineages_corrected,
             "reads_before": number_of_reads,
             "reads_after": number_of_reads_corrected}

    return master_table, stats


def threshold(project, threshold_file):
    thresholds = pl.DataFrame(read_file(threshold_file))
    table = pl.read_parquet(project.hopping_removal_output_path)

    lids_before = len(table)
    reads_before = table.select(pl.col('reads').sum())['reads'][0]

    if len(thresholds.columns) == 1:
        table = table.filter(pl.col('reads') >= thresholds['threshold'][0])
    else:
        columns = table.columns
        table = table.join(thresholds, on=[col for col in table.columns if col != 'threshold' and col in thresholds.columns]).filter(pl.col('reads') >= pl.col('threshold'))
        table = table.select(columns)

    lids_after = len(table)
    reads_after = table.select(pl.col('reads').sum())['reads'][0]

    stats = {"lids_before": lids_before,
             "lids_after": lids_after,
             "reads_before": reads_before,
             "reads_after": reads_after}
    return table.to_arrow(), stats





import pandas
import json
import hashlib
import os.path
import pyarrow
import pyarrow.parquet
from collections import Counter
from phantombuster.io_ import write_parquet
import uuid

def save(result, dir, id=None):
    table, stats  = result

    if id is None:
        id = str(uuid.uuid4())

    if table is not None:
        write_parquet(table, os.path.join(dir, id+'.parquet'))
        r = True
    else:
        r = False

    with open(os.path.join(dir, id + f"_stats.json"), mode="w") as f:
        json.dump(stats, f)

    return (id, r)

def load(id_r, dir):
    id, r = id_r

    if r is True:
        table = pyarrow.parquet.read_table(os.path.join(dir, id+'.parquet'))
    else:
        table = None

    with open(os.path.join(dir, id + f"_stats.json")) as f:
        stats = json.load(f)

    return (table, stats)


def _to_array(iterator, size, type):
    pyarrow.array(iterator, size=size, type=type)

def deduplicator_to_pyarrow_table(deduplicator, columns, lengths, types):
    size = len(deduplicator)
    rows = ((*items, readcount) for items, readcount in deduplicator.items())
    arrays = zip(*rows)
    t = pyarrow.Table.from_arrays([_to_array(iterator, size, type) for iterator, type in zip(arrays, types)], columns)
    return t




import sys
from typing import Coroutine, Callable, Any
from collections.abc import Iterable
import trio

async def gather(*tasks):

    async def collect(index, task, results):
        results[index] = await task

    results = {}
    async with trio.open_nursery() as nursery:
        for index, task in enumerate(tasks):
            nursery.start_soon(collect, index, task, results)
    return [results[i] for i in range(len(tasks))]


async_run = trio.run
async_sleep = trio.sleep
async_gather = gather




import json
import zmq
import trio

class Socket:

    _waiting_time = 0.1

    def __init__(self, ctx=None, **kwargs):
        if ctx is None:
            ctx = zmq.Context.instance()
        self.ctx = ctx
        self._socket = self._make_socket()

    def connect(self, addr):
        return self._socket.connect(addr)

    def listen(self, addr):
        return self._socket.bind(addr)

    def recv(self):
        msg = self._socket.recv()
        return msg.decode("utf8")

    async def async_recv(self):
        while True:
            try:
                msg = self._socket.recv(zmq.DONTWAIT)
                return msg.decode("utf8")
            except zmq.error.Again:
                await trio.sleep(self._waiting_time)

    def recv_json(self):
        msg = self.recv()
        return json.loads(msg)

    async def async_recv_json(self):
        msg = await self.async_recv()
        return json.loads(msg)

    def send(self, msg):
        return self._socket.send(msg.encode("utf8"))

    async def async_send(self, msg):
        return self.send(msg)

    def send_json(self, msg):
        msg_serialized = json.dumps(msg)
        r = self.send(msg_serialized)
        return r

    async def async_send_json(self, msg):
        self.send_json(msg)

    def close(self):
        return self._socket.close()


class PushSocket(Socket):
    def _make_socket(self):
        return self.ctx.socket(zmq.PUSH)


class PullSocket(Socket):
    def _make_socket(self):
        return self.ctx.socket(zmq.PULL)


class ReqSocket(Socket):
    def _make_socket(self):
        return self.ctx.socket(zmq.REQ)


class RepSocket(Socket):
    def _make_socket(self):
        return self.ctx.socket(zmq.REP)



# cython: language_level=3
# distutils: language = c++
# cython: language=c++

import pyarrow
cimport pyarrow
from pyarrow cimport CTable
from pyarrow.lib cimport pyarrow_wrap_array, pyarrow_unwrap_table, TableBatchReader, CRecordBatch, CStringArray, CDoubleArray, CInt64Builder, CArray, c_get_memory_pool, CDoubleBuilder, CUInt64Array, CUInt64Builder, CFixedSizeBinaryArray, pyarrow_wrap_schema, CSchema, Type, CFixedSizeBinaryType, CDataType, CLargeStringArray, CLargeBinaryArray, CBinaryArray

import numpy

cimport cython
from cython.operator cimport dereference as deref

from libcpp.memory cimport shared_ptr, make_shared
from libcpp.string cimport string
from libcpp.cast cimport dynamic_cast, static_cast
from libc.stdint cimport uint8_t, int32_t, int64_t


ctypedef CArray* CArrayPtr
ctypedef CStringArray* CStringArrayPtr
ctypedef CLargeStringArray* CLargeStringArrayPtr
ctypedef CDoubleArray* CDoubdynamicyPtr
ctypedef CUInt64Array* CUInt64ArrayPtr
ctypedef CFixedSizeBinaryArray* CFixedSizeBinaryArrayPtr
ctypedef CFixedSizeBinaryType* CFixedSizeBinaryTypePtr
ctypedef CBinaryArray* CBinaryArrayPtr
ctypedef CLargeBinaryArray* CLargeBinaryArrayPtr

#@cython.boundscheck(False)
#@cython.wraparound(False)
#cdef int _compare(const uint8_t* a, const uint8_t* b):
    #return ((deref(a) > deref(b)) - (deref(a) < deref(b)));
    #if l1 < l2:
        #return -1
    #elif l1 > l2:
        #return 1
    #else:
        #return 0
    #return l1.compare(l2)
    #

@cython.boundscheck(False)
@cython.wraparound(False)
cdef int _compare_string(string a, string b):
    cdef int r = a.compare(b)
    if r > 0:
        return 1
    elif r < 0:
        return -1
    else:
        return 0

@cython.boundscheck(False)
@cython.wraparound(False)
cdef int _compare_uint(const uint8_t* a, const uint8_t* b, int lena, int lenb):
    cdef unsigned int i = 0
    while i < lena:
        if a[i] < b[i]:
            return -1
        elif a[i] > b[i]:
            return 1
        i += 1
    return 0

cdef int _compare_type_aware(shared_ptr[CArray] a, shared_ptr[CArray] b, int i, int j):

    cdef Type t1 = deref(a).type_id()
    cdef Type t2 = deref(b).type_id()

    cdef int32_t length_a
    cdef int32_t length_b

    cdef int64_t length64_a
    cdef int64_t length64_b

    cdef unsigned int lena
    cdef unsigned int lenb
    cdef unsigned int offseta
    cdef unsigned int offsetb
    cdef string str_a
    cdef string str_b
    cdef const uint8_t* bin_a
    cdef const uint8_t* bin_b

    if t1 == Type._Type_STRING:
        str_a = dynamic_cast[CStringArrayPtr](a.get()).GetString(i)
        str_b = dynamic_cast[CStringArrayPtr](b.get()).GetString(j)
        return _compare_string(str_a, str_b)
    if t1 == Type._Type_LARGE_STRING:
        str_a = dynamic_cast[CLargeStringArrayPtr](a.get()).GetString(i)
        str_b = dynamic_cast[CLargeStringArrayPtr](b.get()).GetString(j)
        return _compare_string(str_a, str_b)
    if t1 == Type._Type_FIXED_SIZE_BINARY:
        bin_a = dynamic_cast[CFixedSizeBinaryArrayPtr](a.get()).GetValue(i)
        bin_b = dynamic_cast[CFixedSizeBinaryArrayPtr](b.get()).GetValue(j)
        lena = (dynamic_cast[CFixedSizeBinaryTypePtr](deref(a).type().get())).byte_width()
        lenb = (dynamic_cast[CFixedSizeBinaryTypePtr](deref(b).type().get())).byte_width()
        return _compare_uint(bin_a, bin_b, lena, lenb)
    if t1 == Type._Type_BINARY:
        bin_a = dynamic_cast[CBinaryArrayPtr](a.get()).GetValue(i, &length_a)
        bin_b = dynamic_cast[CBinaryArrayPtr](b.get()).GetValue(j, &length_b)
        return _compare_uint(bin_a, bin_b, length_a, length_b)
    if t1 == Type._Type_LARGE_BINARY:
        bin_a = dynamic_cast[CLargeBinaryArrayPtr](a.get()).GetValue(i, &length64_a)
        bin_b = dynamic_cast[CLargeBinaryArrayPtr](b.get()).GetValue(j, &length64_b)
        return _compare_uint(bin_a, bin_b, length64_a, length64_b)
    return -2

@cython.boundscheck(False)
@cython.wraparound(False)
cdef int compare(shared_ptr[CRecordBatch] a, shared_ptr[CRecordBatch] b, unsigned int i, unsigned int j, cols, unsigned int max_i, unsigned int max_j):

    if i == max_i:
        return 1
    if j == max_j:
        return -1

    cdef int r = 0;

    cdef int ci

    for c in cols:

        ci = c

        r = _compare_type_aware(deref(a).column(ci), deref(b).column(ci), i, j)
        #r = _compare(str_a, str_b)
        if r != 0:
            return r
    return r

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef xmerge(table1, table2, unsigned long[:] idx, unsigned long[:] counts, cols):

    cdef unsigned int i = 0
    cdef unsigned int offset_i = 0
    cdef unsigned int j = 0
    cdef unsigned int offset_j = 0

    cdef unsigned int idx_i = 0
    cdef int c

    cdef int reads_column = table1.column_names.index("reads")

    cdef shared_ptr[CTable] t1 = pyarrow_unwrap_table(table1)
    cdef shared_ptr[CTable] t2 = pyarrow_unwrap_table(table2)

    cdef TableBatchReader* tbr1 = new TableBatchReader(deref(t1))
    cdef TableBatchReader* tbr2 = new TableBatchReader(deref(t2))

    cdef shared_ptr[CRecordBatch] batch1
    cdef shared_ptr[CRecordBatch] batch2

    tbr1.ReadNext(&batch1);
    tbr2.ReadNext(&batch2);

    cdef unsigned int max_i = deref(batch1).num_rows()
    cdef unsigned int max_j = deref(batch2).num_rows()
    cdef unsigned int j_offset = len(table1)

    cdef shared_ptr[CSchema] schema = deref(batch1).schema()

    cdef CUInt64ArrayPtr r1 = static_cast[CUInt64ArrayPtr](deref(batch1).column(reads_column).get())
    cdef CUInt64ArrayPtr r2 = static_cast[CUInt64ArrayPtr](deref(batch2).column(reads_column).get())

    while i < max_i or j < max_j:

        c = compare(batch1, batch2, i, j, cols, max_i, max_j)

        if c == 1:
            idx[idx_i] = j + offset_j + j_offset
            counts[idx_i] = r2.Value(j)
            j += 1
            idx_i += 1
        elif c == -1:
            idx[idx_i] = i + offset_i
            counts[idx_i] = r1.Value(i)
            i += 1
            idx_i += 1
        elif c == 0:
            idx[idx_i] = i + offset_i
            counts[idx_i] = r1.Value(i) + r2.Value(j)
            i += 1
            j += 1
            idx_i += 1
        else:
            return idx, counts, idx_i, i, j, -1

        if i != 0 and i == max_i:
            offset_i += i
            i = 0
            tbr1.ReadNext(&batch1);
            if batch1 == NULL:
                max_i = 0
            else:
                max_i = deref(batch1).num_rows()
                r1 = static_cast[CUInt64ArrayPtr](deref(batch1).column(reads_column).get())
        if j != 0 and j == max_j:
            offset_j += j
            j = 0
            tbr2.ReadNext(&batch2);
            if batch2 == NULL:
                max_j = 0
            else:
                max_j = deref(batch2).num_rows()
                r2 = static_cast[CUInt64ArrayPtr](deref(batch2).column(reads_column).get())

    return idx, counts, idx_i, i, j, 0

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef merge(table1, table2, bhc):

    length = len(table1) + len(table2)

    idx = numpy.zeros(length, dtype=numpy.uint)
    counts = numpy.zeros(length, dtype=numpy.uint)

    column_names = table1.column_names

    t1 = table1
    t2 = table2

    column_names.remove("reads")

    _, __, idx_i, i, j, r = xmerge(t1, t2, idx, counts, [table1.column_names.index(bc['name']) for bc in bhc])
    if r != 0:
        return r

    t3 = pyarrow.concat_tables([table1, table2])

    t3 = t3.remove_column(t3.column_names.index("reads"))
    idx = idx[:idx_i]
    counts = counts[:idx_i]

    a = pyarrow.array(idx)

    counts_array = pyarrow.array(counts)

    t3 = t3.take(a)

    t3 = t3.append_column("reads", counts_array)

    return t3



from pathlib import Path
from phantombuster.store import Store
from phantombuster.remoter import Scheduler
from phantombuster.remoter.serialization import register_serialisation


@register_serialisation
class Project:

    def __init__(self, path, tmp_dir=None, data_dir=None, stats_dir=None, scheduler=None):
        self.path = Path(path)
        self.scheduler = scheduler
        self._tmp_dir = tmp_dir
        self._data_dir = data_dir
        self._stats_dir = stats_dir

    def create(self):
        self.path.mkdir(exist_ok=True)

    def _get_server_path(self):
        path = self.tmp_dir / 'phantombuster.ipc'
        return f"ipc://{path}"

    def _get_server_store_path(self):
        path = self.tmp_dir / 'phantombuster.sqlite3'
        return str(path)

    def get_scheduler(self):
        if self.scheduler is None:
            self.scheduler = Scheduler(address=self._get_server_path(), start_server=True, name="scheduler", server_store_path=self._get_server_store_path())
        return self.scheduler

    @property
    def tmp_dir(self):
        if self._tmp_dir is None:
            tmp_dir = self.path / 'tmp'
        else:
            tmp_dir = Path(self._tmp_dir)
        tmp_dir.mkdir(exist_ok=True)
        return tmp_dir

    @tmp_dir.setter
    def tmp_dir(self, value):
        self._tmp_dir = value

    @property
    def data_dir(self):
        if self._data_dir is None:
            data_dir = self.path / 'data'
        else:
            data_dir = Path(self._data_dir)
        data_dir.mkdir(exist_ok=True)
        return data_dir

    @data_dir.setter
    def data_dir(self, value):
        self._data_dir = value

    @property
    def stats_dir(self):
        if self._stats_dir is None:
            stats_dir = self.path / 'stats'
        else:
            stats_dir = Path(self._stats_dir)
        stats_dir.mkdir(exist_ok=True)
        return stats_dir

    @stats_dir.setter
    def stats_dir(self, value):
        self._stats_dir = value

    @property
    def demultiplex_output_path(self):
        return str(self.data_dir / 'demultiplex.parquet')

    @property
    def demultiplex_stats_path(self):
        return str(self.stats_dir / 'demultiplex_statistics.json')

    @property
    def error_correct_output_path(self):
        return str(self.data_dir / 'error_correct.parquet')

    @property
    def error_correct_stats_path(self):
        return str(self.stats_dir / 'error_correct_statistics.json')

    @property
    def hopping_removal_output_path(self):
        return str(self.data_dir / 'hopping_removal.parquet')

    @property
    def hopping_removal_stats_path(self):
        return str(self.stats_dir / 'hopping_removal_statistics.json')

    @property
    def threshold_output_path(self):
        return str(self.data_dir / 'threshold.parquet')

    @property
    def threshold_stats_path(self):
        return str(self.stats_dir / 'threshold_statistics.json')

    @property
    def pvalue_output_path(self):
        return str(self.data_dir / 'pvalues.csv')

    def _to_json_e(self):
        return {'path': self.path, 'tmp_dir': self._tmp_dir, 'stats_dir': self._stats_dir, 'data_dir': self._data_dir}

    @classmethod
    def _from_json_e(cls, d: dict):
        return Project(**d)





from dataclasses import dataclass
from datetime import datetime
from typing import Any, Callable, Dict, List, NoReturn, Optional, Set, Union, cast
from types import TracebackType
import pickle
import base64

from .serialization import error_to_string, string_to_error, serialize_kwargs, deserialize_kwargs, DATETIME_FORMAT


try:
    from typing import Literal
except:

    class MyLiteral:
        def __init__(self, t):
            self.t = t

        def __getitem__(self, *args, **kwargs):
            return self.t

    Literal = MyLiteral(str)  # type: ignore


class RemoterException(Exception):
    pass


## ==== Messages ====
#
# All Messages that can be exchanged by workers and the server


@dataclass
class FreeMsg:
    """Indicates the worker is free for more work"""

    name: str


@dataclass
class OKMsg:
    """Indicates the requested action succeeded

    Used as answer for Lock, Result and Store messages."""

    pass


@dataclass
class FailMsg:
    """Indicates the requested action failed terminally

    Used as answer for Lock, Result and Store messages."""

    name: str
    message: str


@dataclass
class QueryMsg:
    """Query properties of the server like number of workers or tasks"""

    property: str


@dataclass
class QueryAnswerMsg:
    """Answer for a query"""

    property: str


@dataclass
class TaskMsg:
    """Indicates a task to be executed by the worker"""

    name: str
    persister: str
    task_hash: str
    datetime: datetime
    function_module: str
    function_name: str
    function_hash: str
    kwargs: Dict[str, Any]
    failed: bool
    exception: Optional[tuple[BaseException, TracebackType]]
    request: bool
    time: Optional[float]

@dataclass
class FSaveMsg:
    """Saving a function"""
    datetime: datetime
    function_module: str
    function_name: str
    function_hash: str
    function_code: str

@dataclass
class ResultMsg:
    """Indicates the result of the assigned task.

    outdated, currently `Store`s communicate via StoreMsgs"""

    task_hash: str
    datetime: datetime
    function_name: str
    function_module: str
    kwargs: Dict[str, Any]
    worker_name: str
    result: Any = None


@dataclass
class WaitMsg:
    """Indicate the worker that no work is available, but more might follow

    Is used as an answer to a `FreeMsg` or to a `LockMsg`"""

    time: int


@dataclass
class QuitMsg:
    """Indicate the worker to shut down

    is used as an answer to a `FreeMsg`"""

    name: str
    client_type: str

@dataclass
class LockMsg:
    """Asks to acquire a lock"""

    operation: Literal["lock", "unlock"]
    lock: str
    name: str


@dataclass
class StoreMsg:
    """Updates a Store"""

    datetime: datetime
    function_module: str
    function_name: str
    function_hash: str
    kwargs: Dict[str, Any]
    result: Union[Any, bytes]
    task_hash: str
    result_type: str
    failed: bool
    time: float

    def __repr__(self):
        kwargs = str(self.kwargs)
        if len(kwargs) > 20:
            kwargs = kwargs[:10] + "..." + kwargs[-10:]

        return f"StoreMsg(datetime={self.datetime}, function_module={self.function_module}, function_name={self.function_name}, function_hash={self.function_hash}, kwargs={kwargs}, result=..., task_hash={self.task_hash}, result_type={self.result_type}, failed={self.failed}, time={self.time})"


    def __str__(self):
        kwargs = str(self.kwargs)
        if len(kwargs) > 20:
            kwargs = kwargs[:10] + "..." + kwargs[-10:]

        return f"StoreMsg(datetime={self.datetime}, function_module={self.function_module}, function_name={self.function_name}, function_hash={self.function_hash}, kwargs={kwargs}, result=..., task_hash={self.task_hash}, result_type={self.result_type}, failed={self.failed}, time={self.time})"



@dataclass
class ExceptionMsg:
    name: str
    exception: str

@dataclass
class JoinMsg:
    name: str
    client_type: str

# Typing Stuff to make sure all messages are always handled in all code paths

Message     =          Union[QueryMsg, QueryAnswerMsg, OKMsg , FreeMsg, TaskMsg, QuitMsg, LockMsg, WaitMsg, ResultMsg, StoreMsg, FailMsg, ExceptionMsg, JoinMsg, FSaveMsg]

# A serialized message indicates its type by a token, this lists the available tokens.
# Order is the same as the Message Type
MessageType =               ["QUERY", "ANSWER", "OKOK" , "FREE" ,  "TASK", "QUIT",  "LOCK",  "WAIT",  "RESULT",  "STORE",  "FAIL", "EXCEPTION", "JOIN", "FSAVE"]
# Type of MesssageType, used when only the tokens listed in MessageType are accectable
MessageTypeLiteral = Literal["QUERY", "ANSWER", "OKOK" , "FREE" ,  "TASK", "QUIT",  "LOCK",  "WAIT",  "RESULT",  "STORE",  "FAIL", "EXCEPTION", "JOIN", "FSAVE"]

# Can be used to enable exhaustive type checking in mypy, see https://hakibenita.com/python-mypy-exhaustive-checking
def assert_never(x: NoReturn) -> NoReturn:
    raise Exception("This should never happen.")


# ==== De/Serializing Messages ====


def parse_json_to_message(j: Dict[str, Any], store: 'Store') -> Message:
    """Deserialize a json-like dict to a message"""
    if "type" not in j:
        raise Exception("Message had no declared type.")
    t: MessageTypeLiteral = j["type"]
    if t not in MessageType:
        raise Exception(f"Unknown type {t}")
    if t == "FREE":
        if "name" not in j:
            raise Exception("Free message does not contain worker name")
        return FreeMsg(j["name"])
    elif t == "OKOK":
        return OKMsg()
    elif t == "QUERY":
        return QueryMsg(j["property"])
    elif t == "ANSWER":
        return QueryAnswerMsg(j["property"])
    elif t == "FSAVE":
        return FSaveMsg(
            datetime=datetime.strptime(j["datetime"], DATETIME_FORMAT) if j['datetime'] else None,
            function_module=j["function_module"],
            function_name=j["function_name"],
            function_hash=j["function_hash"],
            function_code=j["function_code"],
        )
    elif t == "TASK":
        return TaskMsg(
            name=j['name'],
            persister=j["persister"],
            task_hash=j["task_hash"],
            datetime=datetime.strptime(j["datetime"], DATETIME_FORMAT) if j['datetime'] else None,
            function_module=j["function_module"],
            function_name=j["function_name"],
            function_hash=j["function_hash"],
            kwargs=deserialize_kwargs(j["kwargs"], store),
            failed=j["failed"],
            exception=string_to_error(j["exception"]),
            request=j["request"],
            time=j["time"]
        )
    elif t == "QUIT":
        return QuitMsg(name=j["name"], client_type=j["client_type"])
    elif t == "LOCK":
        return LockMsg(operation=j["op"], lock=j["lock"], name=j["name"])
    elif t == "WAIT":
        return WaitMsg(j["time"])
    elif t == "RESULT":
        return ResultMsg(
            task_hash=j["task_hash"],
            datetime=datetime.strptime(j["datetime"], DATETIME_FORMAT),
            function_module=j["function_module"],
            function_name=j["function_name"],
            kwargs=deserialize_kwargs(j["kwargs"], store),
            worker_name=j["worker_name"],
            result=j["result"],
        )
    elif t == "STORE":
        r = pickle.loads(base64.b64decode(j["result"].encode("utf-8")))
        return StoreMsg(
            datetime=datetime.strptime(j["datetime"], DATETIME_FORMAT),
            function_module=j["function_module"],
            function_name=j["function_name"],
            function_hash=j["function_hash"],
            kwargs=deserialize_kwargs(j["kwargs"], store),
            result=r,
            task_hash=j["task_hash"],
            result_type=j["result_type"],
            failed=j["failed"],
            time=j["time"],
        )
    elif t == "FAIL":
        return FailMsg(name=j["name"], message=j["message"])
    elif t == "EXCEPTION":
        return ExceptionMsg(name=j["name"], exception=j["exception"])
    elif t == "JOIN":
        return JoinMsg(name=j["name"], client_type=j["client_type"])
    else:
        assert_never(t)


def message_to_json(msg: Message) -> Dict[str, Any]:
    """Serialize a message to json-like dict"""
    if isinstance(msg, OKMsg):
        return {"type": "OKOK"}
    elif isinstance(msg, QueryMsg):
        return {"type": "QUERY", "property": msg.property}
    elif isinstance(msg, QueryAnswerMsg):
        return {"type": "ANSWER", "property": msg.property}
    elif isinstance(msg, FreeMsg):
        return {"type": "FREE", "name": msg.name}
    elif isinstance(msg, TaskMsg):
        return {
            "type": "TASK",
            "name": msg.name,
            "persister": msg.persister,
            "task_hash": msg.task_hash,
            "datetime": msg.datetime.strftime(DATETIME_FORMAT) if msg.datetime else None,
            "function_module": msg.function_module,
            "function_name": msg.function_name,
            "function_hash": msg.function_hash,
            "kwargs": serialize_kwargs(msg.kwargs),
            "failed": msg.failed,
            "exception": error_to_string(msg.exception),
            "request": msg.request,
            "time": msg.time
        }
    elif isinstance(msg, FSaveMsg):
        return {
            "type": "FSAVE",
            "datetime": msg.datetime.strftime(DATETIME_FORMAT) if msg.datetime else None,
            "function_module": msg.function_module,
            "function_name": msg.function_name,
            "function_hash": msg.function_hash,
            "function_code": msg.function_code,
        }
    elif isinstance(msg, QuitMsg):
        return {"type": "QUIT", "name": msg.name, "client_type": msg.client_type}
    elif isinstance(msg, LockMsg):
        return {"type": "LOCK", "op": msg.operation, "name": msg.name, "lock": msg.lock}
    elif isinstance(msg, WaitMsg):
        return {"type": "WAIT", "time": msg.time}
    elif isinstance(msg, ResultMsg):
        return {
            "type": "RESULT",
            "task_hash": msg.task_hash,
            "datetime": msg.datetime.strftime(DATETIME_FORMAT),
            "function_module": msg.function_module,
            "function_name": msg.function_name,
            "kwargs": serialize_kwargs(msg.kwargs),
            "worker_name": msg.worker_name,
            "result": msg.result,
        }
    elif isinstance(msg, StoreMsg):
        r = base64.b64encode(pickle.dumps(msg.result)).decode("utf-8")
        return {
            "datetime": msg.datetime.strftime(DATETIME_FORMAT),
            "type": "STORE",
            "function_module": msg.function_module,
            "function_name": msg.function_name,
            "function_hash": msg.function_hash,
            "kwargs": serialize_kwargs(msg.kwargs),
            "result": r,
            "task_hash": msg.task_hash,
            "result_type": msg.result_type,
            "failed": msg.failed,
            "time": msg.time,
        }
    elif isinstance(msg, FailMsg):
        return {"type": "FAIL", "name": msg.name, "message": msg.message}
    elif isinstance(msg, ExceptionMsg):
        return {"type": "EXCEPTION", "name": msg.name, "exception": msg.exception}
    elif isinstance(msg, JoinMsg):
        return {"type": "JOIN", "name": msg.name, "client_type": msg.client_type}
    else:
        assert_never(msg)





#!/usr/bin/env python

import logging
import logging.config
import os
import os.path
import sys
import multiprocessing as mp
import glob
import collections

import ntpath
from dataclasses import dataclass
import json

from phantombuster import porcelain, plumbing, stores
from phantombuster import core
from phantombuster.stores import deduplicator_to_pyarrow_table
from phantombuster.remoter import Worker
from phantombuster.io_ import write_parquet
from phantombuster.project import Project
import click
import click.core
from typing import Optional, List, Mapping
import pyarrow.parquet
import pyarrow.csv

from pathlib import Path

from gettext import gettext


class OrderedGroup(click.Group):
    def __init__(self, name: Optional[str] = None, commands: Optional[Mapping[str, click.Command]] = None, **kwargs):
        super(OrderedGroup, self).__init__(name, commands, **kwargs)
        self.commands = commands or collections.OrderedDict()
        self.main_commands = collections.OrderedDict()
        self.secondary_commands = collections.OrderedDict()

    def list_commands(self, ctx: click.Context) -> Mapping[str, click.Command]:
        return self.commands

    def list_main_commands(self, ctx: click.Context) -> Mapping[str, click.Command]:
        return self.main_commands

    def list_secondary_commands(self, ctx: click.Context) -> Mapping[str, click.Command]:
        return self.secondary_commands

    def add_main_command(self, cmd, name=None):
        """Registers another :class:`Command` with this group.  If the name
        is not provided, the name of the command is used.
        """
        self.add_command(cmd, name)
        name = name or cmd.name
        if name is None:
            raise TypeError("Command has no name.")
        click.core._check_multicommand(self, name, cmd, register=True)
        self.main_commands[name] = cmd

    def add_secondary_command(self, cmd, name=None):
        """Registers another :class:`Command` with this group.  If the name
        is not provided, the name of the command is used.
        """
        self.add_command(cmd, name)
        name = name or cmd.name
        if name is None:
            raise TypeError("Command has no name.")
        click.core._check_multicommand(self, name, cmd, register=True)
        self.secondary_commands[name] = cmd

    def main_command(self, *args, **kwargs):
        """A shortcut decorator for declaring and attaching a command to
        the group. This takes the same arguments as :func:`command` and
        immediately registers the created command with this group by
        calling :meth:`add_command`.

        To customize the command class used, set the
        :attr:`command_class` attribute.

        .. versionchanged:: 8.0
            Added the :attr:`command_class` attribute.
        """
        from click.decorators import command

        if self.command_class is not None and "cls" not in kwargs:
            kwargs["cls"] = self.command_class

        def decorator(f):
            cmd = command(*args, **kwargs)(f)
            self.add_main_command(cmd)
            return cmd

        return decorator

    def secondary_command(self, *args, **kwargs):
        """A shortcut decorator for declaring and attaching a command to
        the group. This takes the same arguments as :func:`command` and
        immediately registers the created command with this group by
        calling :meth:`add_command`.

        To customize the command class used, set the
        :attr:`command_class` attribute.

        .. versionchanged:: 8.0
            Added the :attr:`command_class` attribute.
        """
        from click.decorators import command

        if self.command_class is not None and "cls" not in kwargs:
            kwargs["cls"] = self.command_class

        def decorator(f):
            cmd = command(*args, **kwargs)(f)
            self.add_secondary_command(cmd)
            return cmd

        return decorator

    def format_commands(self, ctx, formatter):
        """Extra format methods for multi methods that adds all the commands
        after the options.
        """
        main_commands = []
        for subcommand in self.list_main_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            # What is this, the tool lied about a command.  Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue

            main_commands.append((subcommand, cmd))

        secondary_commands = []
        for subcommand in self.list_secondary_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            # What is this, the tool lied about a command.  Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue

            secondary_commands.append((subcommand, cmd))



        # allow for 3 times the default spacing
        if len(main_commands) or len(secondary_commands):
            limit = formatter.width - 6 - max(len(cmd[0]) for cmd in main_commands)

            rows = []
            for subcommand, cmd in main_commands:
                help = cmd.get_short_help_str(limit)
                rows.append((subcommand, help))

            if rows:
                with formatter.section(gettext("Main Commands")):
                    formatter.write_dl(rows)

            limit = formatter.width - 6 - max(len(cmd[0]) for cmd in secondary_commands)

            rows = []
            for subcommand, cmd in secondary_commands:
                help = cmd.get_short_help_str(limit)
                rows.append((subcommand, help))

            if rows:
                with formatter.section(gettext("Secondary Commands")):
                    formatter.write_dl(rows)

def configure_logging(outputlog, verbose):
    logging_config = {
                      'version': 1,
                      'formatters':{'default': {'format': "%(asctime)s %(levelname)-8s %(name)-15s %(message)s",
                                                'datefmt': "%Y-%m-%d %H:%M:%S"}},
                      'handlers': {'console': {'class': 'logging.StreamHandler', 'formatter': 'default', 'stream': 'ext://sys.stdout'}},
                      'loggers': {'remoter': {'level': 'WARNING'}},
                      'root': {'handlers': ['console'], 'level': 'INFO'}
                      }
    if outputlog:
        logging_config['handlers']['file'] = {'class': 'logging.FileHandler', 'formatter': 'default', 'filename': outputlog}
        logging_config['root']['handlers'].append('file')
    if verbose:
        logging_config['root']['level'] = 'DEBUG'

    logging.config.dictConfig(logging_config)
    logging.info('Logging configured')


def log_call(function, **kwargs):
    """Log a CLI call with all arguments"""
    logging.info(f"PhantomBuster {function} was called with the following arguments: {kwargs}")


@click.group(cls=OrderedGroup)
@click.version_option(package_name='phantombuster')
@click.option("--verbose", is_flag=True, default=False, help="Enable verbose debugging")
@click.option("-o", "--outputlog", type=click.Path(), help="Output file for logs")
@click.option("--save-results/--no-save-results", type=bool, default=True, help="DEVELOPER OPTION", hidden=True)
def phantombuster(verbose: bool, outputlog: str, save_results: bool) -> None:
    """Bioinformatical tool to remove phantoms from barcode based NGS sequencing data.

    The tool consists of four stages, represented by the four main commands:

    1. demultiplex     - demultiplex into samples and error correct barcodes with known sequences
    2. error-correct   - error correct barcodes with random sequences
    3. hopping-removal - remove barcode combinations that likely originate from index hopping
    4. threshold       - remove barcode combinations with low read count
    """
    configure_logging(outputlog, verbose)


# -- Main Commands -- #


@phantombuster.main_command()
@click.argument("input-file", type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option("--outdir", required=True, type=click.Path(dir_okay=True, file_okay=False, writable=True, readable=True), help='Directory to save all results and temp files')
@click.option("--regex-file", required=True, type=click.Path(exists=True, dir_okay=False), help='CSV file which specifies via regular expressions where barcodes are located')
@click.option("--barcode-file", type=click.Path(exists=True, dir_okay=False), required=True, help='CSV file that specifies all barcodes and their type')
@click.option("--debug", is_flag=True, hidden=True, default=False)
def demultiplex(input_file, regex_file, debug, outdir, barcode_file):
    """
    Demultiplex BAM/FASTA files into parquet files

    INPUT_FILE is a path to a CSV file that lists all input files

    Requires additional worker processe, see 'phantombuster worker'.
    """
    log_call("demultiplex", input_file=input_file, regex_file=regex_file, debug=debug,
             outdir=outdir, barcode_file=barcode_file)
    project = Project(outdir)

    try:
        core.demultiplex(input_file, regex_file, barcode_file, project, debug=debug)
    except Exception as e:
        logging.exception("Pipeline encountered an error. Aborting.")
        raise click.Abort()
    return


@phantombuster.main_command()
@click.option("--outdir", required=True, help='Directory to save all results and temp files')
@click.option("--error-threshold", default=1, help='Maximal Hamming distance to consider two barcode sequences related')
@click.option("--barcode-file", required=True, type=click.Path(exists=True, file_okay=True, dir_okay=False), help='CSV file that specifies all barcodes and their type')
@click.option("--remove-ambigious/--keep-ambigious", default=False, help='Remove combinations with ambigious characters (NRYSWKMBDHV) after error correction')
def error_correct(outdir, error_threshold, barcode_file, remove_ambigious):
    """
    Correct sequencing errors in random barcode sequences originating from single-nucleotide errors

    Requires additional worker processe, see 'phantombuster worker'.
    """
    log_call("error-correct", outdir=outdir, error_threshold=error_threshold, barcode_file=barcode_file)
    project = Project(outdir)
    core.error_correct(project, error_threshold, barcode_file, remove_ambigious)


@phantombuster.main_command()
@click.argument('hopping-barcodes', nargs=-1)
@click.option("--outdir", required=True, type=click.Path(dir_okay=True, file_okay=False), help='Directoy to save all results and temp files')
@click.option("--threshold", default=0.05, type=click.FloatRange(min=0.0, max=1.0), help='p-value threshold. Lower is more strict.')
def hopping_removal(outdir, threshold, hopping_barcodes):
    """
    Remove phantom combinations originating from index hopping

    The read count of each barcode combination is compared to the expected read count under index hopping.
    """
    log_call("hopping-removal", outdir=outdir, threshold=threshold, hopping_barcodes=hopping_barcodes)
    project = Project(outdir)
    hopping_barcodes = [bc.split(',') for bc in hopping_barcodes]
    core.hopping_removal(project, hopping_barcodes, threshold)

@phantombuster.main_command()
@click.option("--outdir", required=True, type=click.Path(dir_okay=True, file_okay=False), help='Directory to save all results and temp files')
@click.option("--threshold-file", required=True, type=click.Path(file_okay=True, dir_okay=False, exists=True), help='CSV file that specifies the read threshold under which combinations are removed')
def threshold(outdir, threshold_file):
    """
    Remove combinations with a read count below a threshold
    """
    log_call("threshold", outdir=outdir, threshold_file=threshold_file)
    project = Project(outdir)
    core.threshold(project, threshold_file)


@phantombuster.main_command()
@click.option("--outdir", required=True, type=click.Path(dir_okay=True, file_okay=False), help='Directory to save all results and temp files')
@click.option("--category-file", required=True, type=click.Path(file_okay=True, dir_okay=False, exists=True), help='CSV file that specifies the categories for each grna')
@click.option("--read-file", required=False, default=None, type=click.Path(file_okay=True, dir_okay=False, exists=True), help='Read file to calculate pvalues on. Defaults to the output file of the thresholding step')
@click.option('--reads-column', default='reads', help='Column used to calculate test statistic with')
@click.option('--group-by', default='grna', help='Defines the columns used to group individual data points (lineages) to units of analysis (grna). Comma seperated values are interpreted as multiple columns (e.g. `grna,sample`)')
@click.option('--N', default=1_000_000, help='Size of distribution used for p-value calculation. Smallest non-zero p-value is 1/N')
@click.option('--max-n', default=950, help='Maximal number of lineages used for variance estimate. Units with more lineages are treated as having max_n lineages.')
def pvalue(outdir, category_file, read_file, reads_column, group_by, n, max_n):
    """
    Calculate pvalues for each gRNA
    """
    log_call("pvalue", outdir=outdir, category_file=category_file, read_file=read_file, reads_column=reads_column, group_by=group_by, N=n, max_n=max_n)
    project = Project(outdir)
    if read_file is None:
        read_file = project.threshold_output_path

    group_by = group_by.split(',')
    core.pvalue(project, read_file, category_file, column=reads_column, group_by=group_by, N=n, max_n=max_n)

# -- Helper Commands -- #

@phantombuster.secondary_command()
@click.option("--outdir", default=None, required=True, type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True))
@click.option("--name", default=None)
def worker(outdir, name):
    """
    Start a worker process

    The worker process uses IPC to connect to the server. It needs to be run on the same node
    and in the same working directory as the server.
    """
    import phantombuster as phantombuster
    import phantombuster.plumbing

    project = Project(outdir)
    project.create()
    path = project._get_server_path()

    worker = Worker(path, name=name)
    worker.start_async()

@phantombuster.secondary_command()
@click.argument("parquetfile", type=click.Path(exists=True, readable=True))
@click.argument("outfile", default=None, required=False, type=click.Path(writable=True))
def to_csv(parquetfile, outfile):
    """
    Convert a parquet file to a CSV file.
    """
    log_call("to_csv", sample=parquetfile, outdir=outfile)
    table = pyarrow.parquet.read_table(parquetfile)
    if outfile is None:
        outfile = parquetfile.replace(".parquet", ".csv")
    pyarrow.csv.write_csv(table, outfile)


@phantombuster.secondary_command()
@click.argument("csvfile", type=click.Path(exists=True, readable=True, dir_okay=False))
@click.argument("outfile", default=None, required=False, type=click.Path(exists=False, writable=True, dir_okay=False))
def to_parquet(csvfile, outfile):
    """
    Convert a CSV file to a parquet file.
    """
    log_call("to_parquet", csvfile=csvfile, outfile=outfile)
    table = pyarrow.csv.read_csv(csvfile)
    if outfile is None:
        outfile = csvfile.replace(".csv", ".parquet")
    write_parquet(table, outfile)


@phantombuster.secondary_command()
@click.argument("prefixes", nargs=-1)
@click.option("--outdir", required=True, type=click.Path(dir_okay=True, file_okay=False, readable=True, writable=True))
@click.option("--prefix")
@click.option("--barcode-hierarchy-file", type=click.Path(exists=True), required=True)
def merge(prefixes, outdir, prefix, barcode_hierarchy_file):
    """
    Merge multiple prefixes under one prefix
    """
    # Log the call of this function with all parameters to the logfile
    log_call("merge", prefixes=prefixes, outdir=outdir, prefix=prefix, barcode_hierarchy_file=barcode_hierarchy_file)

    master_paths = PathsAndFiles(outdir, prefix, None)
    master_paths.create()

    try:
        barcode_hierarchy = plumbing.read_barcode_hierarchy_file(barcode_hierarchy_file)
    except Exception:
        raise Exception("Could not read barcode hierarchy file correctly")

    to_merge = [PathsAndFiles(outdir, prefix, None) for prefix in prefixes]

    results = [stores.load(('deduplication', True), paths.stage_path('deduplication')) for paths in to_merge]
    out = plumbing.combine(results, barcode_hierarchy)

    stores.save(out, master_paths.stage_path('deduplication'), id='deduplication')

if __name__ == "__main__":
    phantombuster()



from typing import Optional
from types import TracebackType
from phantombuster.config_files import InputFile, RegexDictionary, InputGroup

import pickle
import json

DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"

def error_to_string(err: Optional[tuple[BaseException, TracebackType]]) -> Optional[str]:
    if err:
        return str(err)
        #return pickle.dumps(err, protocol=pickle.HIGHEST_PROTOCOL).decode("utf8", "surrogateescape")
    return None

def string_to_error(s: Optional[str]) -> Optional[tuple[BaseException, TracebackType]]:
    if s:
        return s, None
        return pickle.loads(s.encode("utf8", "surrogateescape"))
    return None

EXTENSIONS = {}

def register_serialisation(cls):
    EXTENSIONS[cls.__name__] = cls
    return cls

def register_serialization_extension(name, cls):
    EXTENSIONS[name] = cls

register_serialization_extension('InputFile', InputFile)
register_serialization_extension('RegexDictionary', RegexDictionary)
register_serialization_extension('InputGroup', InputGroup)

def serialize_value(value):
    from .task import Task
    if isinstance(value, Task):
        value.calc_hash()
        return ('t', value.task_hash)
    if isinstance(value, list):
        return ('l', [serialize_value(v) for v in value])
    if isinstance(value, dict):
        return ('d', {key: serialize_value(v) for key, v in value.items()})

    for name, cls in EXTENSIONS.items():
        if isinstance(value, cls):
            return ('e', {'name': name, 'value': value._to_json_e()})
    return ('a', value)

def deserialize_value(value, store):
    value_type, value = value
    if value_type == 't':
        task = store.load_from_hash(value)
        if task is None:
            raise Exception(f'Could not deserialize task with hash {value["value"]}')
        return task
    elif value_type == 'l':
        return [deserialize_value(v, store) for v in value]
    elif value_type == 'd':
        return {key: deserialize_value(v, store) for key, v in value.items()}
    elif value_type == 'e':
        cls = EXTENSIONS[value['name']]
        return cls._from_json_e(value['value'])
    elif value_type == 'a':
        return value


def serialize_kwargs(kwargs):
    if kwargs is not None:
        kwargs = {key: serialize_value(value) for key, value in kwargs.items()}
        kwargs = _clean(kwargs)
    return json.dumps(kwargs)

def deserialize_kwargs(string, store):
    kwargs = json.loads(string)
    if kwargs is not None:
        kwargs = {key: deserialize_value(value, store) for key, value in kwargs.items()}
    return kwargs

def _clean(v):
    if isinstance(v, dict):
        r = {}
        for key in sorted(v):
            r[key] = _clean(v[key])
        return r
    return v




from math import floor, ceil
from collections import defaultdict, namedtuple, Counter
import sys
import logging


# Immutable representation of an UMI
Umi = namedtuple("Umi", ["start", "end", "tag"])
# A UMI, consisting of start- and end-position and tag
Result = namedtuple("Result", ["ecumi_counts", "ecumi_list"])

def hamming_distance(s1: str, s2: str) -> int:
    """Computed the (unscaled) hamming distance between two lists"""
    if len(s1) != len(s2):
        raise RuntimeError(
            "hamming_distance required sequences with the same length, but give %d and %d characters"
            % (len(s1), len(s2))
        )
    return sum(ch1 != ch2 for ch1, ch2 in zip(s1, s2))


    def increment(self):
        self.value.value += 1

    def get_value(self):
        return self.value.value


class UmiEC:
    """
    A mutable node in a UMI similarity graph/forest.

    Mutable representation of a node in the UMI forest. The field 'umi' contains
    an Umi, and 'parents' a set of parent nodes, which the Umi is similar too,
    *and* which have a higher total read count. If the total read counts of two
    similar Umis are equal, the less-than (<) operator of the Umi type is used to
    break the tie, and the "greater" Umi is made the parent. This ensures
    non-circularity of the graph in all cases.
    """

    def __init__(self, umi, parents):
        self.umi = umi
        self.parents = parents

    # Make hashable, use object identity as key (as is appropriate for a mutable type)
    def __hash__(self):
        return id(self)

    # Identity means object identity (see __hash__ above)
    def __eq__(self, other):
        return id() == other.id()


class UmiCount:
    """ "
    Per-UMI read count for data without stranded UMIs

    Tracks not only the total read count, but also the number of UMIs merged into
    one during error-correction (`rawumis`).
    """

    @classmethod
    def zero(cls):
        return cls(0, 0)

    @classmethod
    def rawumi(cls):
        return cls(0, 1)

    def __init__(self, count, rawumis):
        self.rawumis = rawumis
        self.count = count

    def total(self):
        return self.count

    def __add__(self, other):
        return UmiCount(self.count + other.count, self.rawumis + other.rawumis)

    def __eq__(self, other):
        return self.count == other.count

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return str(self.count)

    def __repr__(self):
        return str(self.count)


class UmiCountStranded:
    """ "
    Per-UMI read count for stranded UMIs, i.e. if both strands are detectable

    Tracks the read counts of both strands separately, and also the number of UMIs
    merged into one during error-correction (`rawumis`).
    """

    @classmethod
    def zero(cls):
        return cls(0, 0, 0)

    @classmethod
    def rawumi(cls):
        return cls(0, 0, 1)

    def __init__(self, plus, minus, rawumis):
        self.rawumis = rawumis
        self.plus = plus
        self.minus = minus

    def total(self):
        return self.plus + self.minus

    def __add__(self, other):
        return UmiCountStranded(
            self.plus + other.plus,
            self.minus + other.minus,
            self.rawumis + other.rawumis,
        )

    def __eq__(self, other):
        return (self.plus == other.plus) and (self.minus == other.minus)

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return "%d+,%d-" % (self.plus, self.minus)

class ErrorCorrector:
    """
    UMI error correction.

    Two Umis whose start and end-positions each differ by at most max_mapping_distance
    bases, and whose tags have hamming distance at most max_hamming_distance are merged
    into one.
    """

    def __init__(
        self,
        tag_length,
        max_mapping_distance_start,
        max_mapping_distance_end,
        max_hamming_distance,
        count_type=UmiCount,
        report=False,
    ):
        self.count_type = count_type
        self.tag_length = tag_length
        self.max_mapping_distance_start = max_mapping_distance_start
        self.max_mapping_distance_end = max_mapping_distance_end
        self.max_hamming_distance = max_hamming_distance
        self.tsw = int(floor(tag_length / (self.max_hamming_distance + 1)))
        self.report = report
        self.report_step_umis = 500
        self.report_step_basepairs = 50

    def process(self, umi_counts):
        if self.report:
            logging.debug(
                "%d UMIs with %s reads before error-correction\n",
                 len(umi_counts), sum(umi_counts.values(), self.count_type.zero())
            )

        # Build forest
        ecinfo = self.build_forest(umi_counts)
        # Find roots and sum counts
        result = self.sum_trees(umi_counts, ecinfo)
        ecumi_counts = result.ecumi_counts

        if self.report:
            sys.stderr.write(
                "%d UMIs with %s reads after error-correction\n"
                % (
                    len(ecumi_counts),
                    sum(ecumi_counts.values(), self.count_type.zero()),
                )
            )

        return result

    def link_if_similar(self, umi, umi2, umi_counts, ecinfo):
        # Debugging support.
        log = False
        if log:
            logging.debug("UMI1: %s x %s", umi, umi_counts[umi])
            logging.debug("UMI2: %s x %s", umi2, umi_counts[umi2])
        # Check mapping distance (start)
        if abs(umi2.start - umi.start) > self.max_mapping_distance_start:
            # Mapping positions lie too far apart to unify UMIs.
            if log:
                logging.debug("DEBUG: NOT MERGED, start distance > %d", self.max_mapping_distance_start)
            return
        elif umi2.start < umi.start:
            # Impossible, list is sorted!
            raise RuntimeError("internal error, UMI list corrupted")
        # Check mapping distance (end)
        if abs(umi2.end - umi.end) > self.max_mapping_distance_end:
            # Mapping positions lie too far apart to unify UMIs
            if log:
                logging.debug("NOT MERGED, end distance > %d", self.max_mapping_distance_end)
            return
        # Check hamming distance
        hd = False
        if self.max_hamming_distance > 0:
            hd = hamming_distance(umi.tag, umi2.tag) > self.max_hamming_distance
        else:
            hd = umi.tag != umi2.tag
        if hd:
            # Hamming distance too large to allow unification of UMIs
            if log:
                logging.debug("DEBUG: NOT MERGED, hamming distance > %d", self.max_hamming_distance)
            return
        # Determine which to merge into which
        umi_cnt = umi_counts[umi].total()
        umi2_cnt = umi_counts[umi2].total()
        if (umi_cnt, umi) < (umi2_cnt, umi2):
            p, c = umi2, umi
        elif (umi2_cnt, umi2) < (umi_cnt, umi):
            p, c = umi, umi2
        else:
            raise RuntimeError("Tertium non datur fails for %s and %s" % (umi, umi2))
        if log:
            logging.debug("WILL MERGE %s x %s --> %s x %s", c, umi_counts[c], p, umi_counts[p])
        # Merge UMIs as described above
        ecinfo[c].parents.add(ecinfo[p])

    # Splits a tag into enough disjoint parts such that whenever the hamming
    # distance of two tags is at most max_hamming_distance, then at least one
    # part is identical in both tags.
    def splittag(self, tag):
        if len(tag) != self.tag_length:
            raise ValueError("tag %s has wrong length, %d instead of %d" % (tag, len(tag), self.tag_length))
        return [
            tag[i * self.tsw : ((i + 1) * self.tsw if i < self.max_hamming_distance else self.tag_length)]
            for i in range(self.max_hamming_distance + 1)
        ]

    def build_forest(self, umi_counts):
        if self.report:
            logging.debug("Finding similar UMIs...\n")

        # Data-structure which represents the forest. ecinfo associates each Umi
        # with an UmiEC instance, which contains the Umi itself, and a set of
        # parents. Note that since parents is a set, the "forest" does not actually
        # contain trees, but DAGs.
        ecinfo = dict()
        for u in umi_counts.keys():
            ecinfo[u] = UmiEC(u, set())

        # Build Umi index. The first level splits Umis by their starting position.
        # Below that, there are max_hamming_distance + 1 associative maps (dicts),
        # M_k, one for each of the parts returned by splittag(). The map M_k maps a
        # string s to the set of Umis whose k-th tag part equals s. Therefore, given
        # a tag x, all tags y whose hamming distance is at most max_hamming_distance,
        # are contained in at least one of the sets M_k(x_k), where x_k is the k-th
        # tag part of x.
        umi_index = defaultdict(lambda: [defaultdict(set) for i in range(0, self.max_hamming_distance + 1)])
        for umi in umi_counts.keys():
            t = umi_index[umi.start]
            for i, p in enumerate(self.splittag(umi.tag)):
                t[i][p].add(umi)

        # Populate ecinfo
        last_pos = 0
        last_report_pos = 0
        # Scan Umis
        for i, umi in enumerate(sorted(umi_counts.keys(), key=lambda u: u.start)):
            if umi.start < last_pos:
                raise RuntimeError("UMIs not sorted by start position (%d after %d)!" % (umi.start, last_pos))
            if self.report and (i % self.report_step_umis == 0):
                logging.debug("  processed %d UMIs\n", i)
            if self.report and ((umi.start - last_report_pos) >= self.report_step_basepairs):
                last_report_pos = umi.start
                logging.debug("  processed %d basepairs\n", umi.start)
            last_pos = umi.start
            # Scan part of first index level corresponding to a similar start position.
            for s2 in range(umi.start, umi.start + self.max_mapping_distance_start + 1):
                # Scan tag parts of 1st Umi
                for k2, p2 in enumerate(self.splittag(umi.tag)):
                    # Scan Umis having the same k2-th tag part p2 as the 1st Umi
                    for umi2 in umi_index[s2][k2][p2]:
                        # Avoid calling link_if_similar twice for every pair, using the total
                        # order imbued on the Umis to decide which call to suppress. Note that
                        # whatever condition we use here *must* ensure that if umi has a lower
                        # start position than umi2, link_if_similar *will* be called, since we
                        # only let s2 range within [ umi.start, umi.start + max_dist ] above.
                        if umi < umi2:
                            self.link_if_similar(umi, umi2, umi_counts, ecinfo)

        # Return forest
        return ecinfo

    def sum_trees(self, umi_counts, ecinfo):
        if self.report:
            logging.debug("Merging similar UMIs...\n")

        # All counts within an UMI tree contribute to the roots count.
        # If the tree is actually a net, i.e. if there is more than one "root"
        # reachable from a certain leaf, that leaf's assignment is ambiguous,
        # and its count is not attributed to any root.
        processed = 0
        ambiguous = 0
        ambiguous_count = self.count_type.zero()
        ecumi_counts = defaultdict(lambda: self.count_type.zero())
        ecumi_list = {}
        for umi, umi_ec in ecinfo.items():
            if self.report and (processed % self.report_step_umis == 0):
                logging.debug("  processed %d UMIs\n" % processed)
            processed += 1
            # Start with current node as "active set"
            nodes = set([umi_ec])
            roots = []
            while nodes:
                # Pick node in active set,
                n = nodes.pop()
                if n.parents:
                    # and replace it by its parents
                    nodes.update(set(n.parents))
                else:
                    roots.append(n.umi)
            if len(roots) == 1:
                # Found a unique root, so add node's counts to root's counts
                if umi_counts[umi].rawumis != 1:
                    raise RuntimeError("before error correction, umis should have rawumis=1")
                ecumi_counts[roots[0]] += umi_counts[umi]
                ecumi_list.update({umi: roots[0]})
            elif len(roots) > 1:
                # Found more than one root, so throw node away
                if umi_counts[umi].rawumis != 1:
                    raise RuntimeError("before error correction, umis should have rawumis=1")
                ambiguous_count += umi_counts[umi]
            else:
                ecumi_list.update({umi: umi})
        # Check that we haven't missed any UMI
        if sum(umi_counts.values(), self.count_type.zero()) != (
            sum(ecumi_counts.values(), self.count_type.zero()) + ambiguous_count
        ):
            raise RuntimeError("error correction buggy!")
        if self.report:
            logging.debug("%d UMIs with %s reads were ambiguous\n", ambiguous_count.rawumis, ambiguous_count)

        result = Result(ecumi_counts, ecumi_list)
        return result



from typing import Any, Callable, Optional
import os.path
import sys
import inspect
from datetime import datetime
import logging
import uuid
import time
import threading

from . import globaladdress
from .persisters import PersisterMapper
from .task import Task, hash_function
from .store import Store
from .messages import Message, StoreMsg, JoinMsg, WaitMsg, QuitMsg, parse_json_to_message, message_to_json, QueryMsg, QueryAnswerMsg
from .async_ import async_gather, async_sleep
from .socket_ import Socket, ReqSocket
from .worker import Worker
from .server import Server

class ServerManager:

    def __init__(self, address, storepath=None):
        self.address = address
        self.server = None
        self.storepath = storepath

    def start(self):
        self.thread = threading.Thread(target=self._start)
        self.thread.start()

    def _start(self):
        self.server = Server(self.address, storepath=self.storepath)
        self.server.start_async()

    def wait(self):
        self.thread.join()

    def stop(self):
        self.thread.stop()


class SingleWorker:

    def __init__(self, address, name):
        self.address = address
        self.name = name

    def start(self):
        self.thread = threading.Thread(target=self._start)
        self.thread.start()

    def _start(self):
        self.worker = Worker(self.address, self.name)
        self.worker.start_async()

    def stop(self):
        self.thread.stop()

    def wait(self):
        self.thread.join()


class WorkerManager:

    def __init__(self, address):
        self.address = address
        self.workers = []

    def start(self):
        worker = SingleWorker(self.address, name=f'worker{len(self.workers)+1}')
        self.workers.append(worker)
        worker.start()

    def wait(self):
        for worker in self.workers:
            worker.wait()

    def stop(self):
        for worker in self.workers:
            worker.stop()


class Scheduler:

    DEFAULT_WORKERS = 0

    def __init__(self,
                 address: Optional[str] = None,
                 persister_mapper=PersisterMapper,
                 name: Optional[str] = None,
                 start_server: bool = False,
                 start_workers: int = -1,
                 save_functions: bool = False,
                 server_store_path: Optional[str] = None,
                 ):
        # If no address is given and no explicit address was ever given, use default address
        if address is None and globaladdress.global_address is None:
            address = "ipc://.remoter.ipc"
        # if explicit address was given, reuse it for later
        elif address is not None and globaladdress.global_address is None:
            globaladdress.global_address = address
        # if no explicit address is given, reuse last explicit address
        elif address is None and globaladdress.global_address is not None:
            address = globaladdress.global_address

        if start_workers == -1:
            start_workers = Scheduler.DEFAULT_WORKERS

        self.worker_manager = WorkerManager(address)
        self.server_manager = ServerManager(address, server_store_path)

        self._address = address
        if name:
            self._name = name
        else:
            self._name = uuid.uuid4().hex
        self._running = False
        self._persister_mapper = persister_mapper
        self._current_file_module = os.path.split(sys.argv[0])[-1].replace(".py", "")
        self._store = Store(":memory:", name=self._name)
        self._logger = logging.getLogger(f"remoter.scheduler.{self._name}")
        self._logger.debug(f"Connecting to {address}")
        self._socket: Socket = self._create_socket()
        self._socket.connect(address)
        self._joined = False
        self._entered = 0
        self._failed = False
        self._tasks: list["Task"] = []
        self._start_server = start_server
        self._start_workers = start_workers
        self._save_functions = save_functions




    def _create_task(self, f: Callable, *args: Any, **kwargs: Any) -> Task:

        params = inspect.signature(f).parameters
        kwargs.update({param: arg for param, arg in zip(params, args)})
        module = f.__module__ if f.__module__ != "__main__" else self._current_file_module
        code = inspect.getsource(f)
        task = Task(None, self._store, datetime.now(), module, f.__name__, hash_function(f), kwargs, code=code)
        task.persister = self._persister_mapper(f)
        task.task_hash = task.calc_hash()
        return task

    def _fullfill(self, task, result):
        task.result = result
        task.done = True
        task._save_force()

    def _create_socket(self):
        return ReqSocket()

    def _join(self):
        msg = JoinMsg(self._name, "scheduler")
        self._send(msg)
        msg = self._recv()
        self._joined = True

    def run(self, f: Callable, *args: Any, **kwargs: Any) -> Task:
        if not self._joined:
            self._join()
        task = self.schedule(f, *args, **kwargs)
        self.gather(task)
        return task.result

    def request(self, f: Callable, *args: Any, **kwargs: Any) -> Any:
        if not self._joined:
            self._join()
        task = self._create_task(f, *args, **kwargs)
        self._logger.info("scheduling task: %s", task.function_name)
        msg = task.to_task_message(name=self._name, request=False)
        self._send(msg)
        return_msg = self._recv()
        if isinstance(return_msg, StoreMsg):
            self._store._income(return_msg)
            task.calc_hash()
            task.load()
        return task

    def schedule(self, f: Callable, *args: Any, **kwargs: Any) -> Any:
        if not self._joined:
            self._join()
        task = self._create_task(f, *args, **kwargs)
        if self._save_functions:
            msg = task.to_fsave_message()
            self._send(msg)
            return_msg = self._recv()
        self._logger.info("scheduling task: %s", task.function_name)
        msg = task.to_task_message(name=self._name)
        self._send(msg)
        return_msg = self._recv()
        if isinstance(return_msg, StoreMsg):
            self._store._income(return_msg)
            task.calc_hash()
        self._tasks.append(task)
        return task

    def wait_all(self, tasks, unpack=False):
        return [self.wait(task, unpack=unpack) for task in tasks]

    def wait(self, task, unpack=False, return_after_wait=False):
        return self._wait_for(task, unpack=unpack, return_after_wait=return_after_wait)

    def _wait_for(self, task, unpack=False, return_after_wait=False):
        self._logger.debug("Waiting for completion of task: %s", task)
        task.calc_hash()
        while not task.is_loadable():
            msg = task.to_task_message(self._name)
            self._send(msg)
            msg = self._recv()

            if isinstance(msg, StoreMsg):
                self._store._income(msg)
            elif isinstance(msg, WaitMsg):
                time.sleep(msg.time)
                if return_after_wait:
                    return msg.time
            elif isinstance(msg, QuitMsg):
                self._failed = True
                self._joined = False
                return None
            else:
                raise Exception(f"Scheduler received unexpected message: {msg}")
        self._logger.debug("Successfully waited for completion of task: %s", task)
        if unpack:
            task.calc_hash()
            task.load()
            return task.result
        else:
            return task

    def query_workers(self):
        self._logger.debug('Quering for number of workers')
        msg = QueryMsg("workers")

        answered = False

        workers = None

        while not answered:
            self._send(msg)
            msg = self._recv()

            if isinstance(msg, WaitMsg):
                time.sleep(msg.time)
            elif isinstance(msg, QuitMsg):
                self._failed = True
                self._joined = False
                return None
            elif isinstance(msg, QueryAnswerMsg):
                answered = True
                if msg.property != "None":
                    workers = int(msg.property)

        return workers

    def gather(self, *tasks):
        if self._joined:
            for task in tasks:
                self._wait_for(task)
            rs = [task.load() for task in tasks]
            for task in tasks:
                if task.failed:
                    err, tb = task.exception
                    raise err.with_traceback(tb)
            return rs
        else:
            if all(task.is_loadable() for task in tasks):
                rs = [task.load() for task in tasks]
                for task in tasks:
                    if task.failed:
                        err, tb = task.exception
                        raise err.with_traceback(tb)
                return rs
            else:
                raise Exception("Not all tasks are loadable and I'm not connected to a server")

    def _recv(self) -> Message:
        if self._socket is not None:
            self._logger.debug("Waiting for message sync")
            msg_json = self._socket.recv_json()
            msg = parse_json_to_message(msg_json, self._store)
            self._logger.debug(f"Sync Received message: {msg}")
            return msg
        else:
            raise Exception("Scheduler has to connect before being able to receive messages")

    def _send(self, msg: Message) -> None:
        if self._socket:
            j = message_to_json(msg)
            self._logger.debug(f"Sync Sending message: {msg}")
            self._socket.send_json(j)
        else:
            raise Exception("Scheduler has to connect before being able to send messages")

    def __enter__(self):
        if self._entered == 0:
            if self._start_server:
                self.server_manager.start()
            for i in range(self._start_workers):
                self.worker_manager.start()
            self._join()
        self._entered += 1
        return self

    def __exit__(self, exc_type, exc, tb):
        if self._entered == 1:
            self._logger.debug("%s, %s, %s", exc_type, exc, tb)
            self._logger.debug("Scheduler exited, sending exit msg to server sync")
            self.exit(force=True)
            self._logger.debug("Notified server of exit, exiting now")
            self.server_manager.wait()
            self.worker_manager.wait()
        self._entered -= 1
        if exc:
            raise exc
        return self

    def exit(self, force=False):
        if not force:
            for task in self._tasks:
                self._wait_for(task)
        self._send(QuitMsg(self._name, "scheduler"))
        self._recv()
        self._joined = False
        self._running = False

    async def _async_join(self):
        msg = JoinMsg(self._name, "scheduler")
        await self._async_send(msg)
        msg = await self._async_recv()
        self._joined = True

    async def async_run(self, f: Callable, *args: Any, **kwargs: Any) -> Any:
        if not self._joined:
            await self._async_join()
        task = await self.async_schedule(f, *args, **kwargs)
        await task.wait()
        return task.result

    async def async_schedule(self, f: Callable, *args: Any, **kwargs: Any) -> Any:
        if not self._joined:
            await self._async_join()
        task = self._create_task(f, *args, **kwargs)
        if self._save_functions:
            msg = task.to_fsave_message()
            await self._async_send(msg)
            return_msg = await self._async_recv()
        self._logger.info("scheduling task: %s", task.function_name)
        msg = task.to_task_message(self._name)
        await self._async_send(msg)
        return_msg = await self._async_recv()
        if isinstance(return_msg, StoreMsg):
            self._store._income(return_msg)
            task.calc_hash()
            task.load()
        self._tasks.append(task)
        return task

    async def _async_wait_for(self, task):
        self._logger.debug("Waiting for completion of task: %s", task)
        task.calc_hash()

        while not task.is_loadable():
            msg = task.to_task_message(self._name)
            j = message_to_json(msg)
            self._logger.debug(f"Sending message: {msg}")
            await self._socket.async_send_json(j)

            self._logger.debug("Waiting for message async")
            msg_json = await self._socket.async_recv_json()
            msg = parse_json_to_message(msg_json, self._store)
            self._logger.debug(f"Received message: {msg}")

            if isinstance(msg, StoreMsg):
                self._store._income(msg)
                task.calc_hash()
                task.load()
            elif isinstance(msg, WaitMsg):
                await async_sleep(msg.time)
            elif isinstance(msg, QuitMsg):
                self._failed = True
                self._joined = False
                return None
            else:
                raise Exception(f"Scheduler received unexpected message: {msg}")
        self._logger.debug("Successfully waited for completion of task: %s", task)
        return task.result

    async def async_gather(self, *tasks):
        rs = []
        for task in tasks:
            r = await self._async_wait_for(task)
            rs.append(r)
        for task in tasks:
            if task.failed:
                raise Exception(task.result)
        return rs

    async def _async_recv(self) -> Message:
        if self._socket is not None:
            self._logger.debug("Waiting for message async")
            msg_json = await self._socket.async_recv_json()
            msg = parse_json_to_message(msg_json, self._store)
            self._logger.debug(f"Received message: {msg}")
            return msg
        else:
            raise Exception("Scheduler has to connect before being able to receive messages")

    async def _async_send(self, msg: Message) -> None:
        if self._socket:
            j = message_to_json(msg)
            self._logger.debug(f"Sending message: {msg}")
            await self._socket.async_send_json(j)
        else:
            raise Exception("Scheduler has to connect before being able to send messages")

    async def __aenter__(self):
        if self._entered == 0:
            if self._start_server:
                start_server(self._address)
            for i in range(self._start_workers):
                start_worker(self._address, i)
            self._logger.debug("Scheduler entered async, sending join msg to server async")
            await self._async_join()
            self._logger.debug("Joined async")
        self._entered += 1
        return self

    async def __aexit__(self, exc_type, exc, tb):
        if self._entered == 1:
            self._logger.error("%s, %s, %s", exc_type, exc, tb)
            self._logger.debug("Scheduler exited async, sending exit msg to server async")
            await self.async_exit()
            self._logger.debug("Notified server of exit, exiting now")
        self._entered -= 1
        if exc:
            raise exc
        return self

    async def async_exit(self):
        rs = []
        for task in self._tasks:
            r = await self._async_wait_for(task)
            rs.append(r)
        await self._async_send(QuitMsg(self._name, "scheduler"))
        await self._async_recv()
        self._joined = False
        self._running = False
        return rs



from functools import wraps
from hashlib import sha3_256
import inspect
from datetime import datetime
from typing import Any
import uuid
import sqlite3

from dataclasses import dataclass, field
import dataclasses

from phantombuster import store
from phantombuster.store import Store, StoreID

import pickle
import time

from typing import Any, Optional, Union, Callable, Dict, List

@dataclass
class FunctionDefinition:
    name: str
    source: str
    function_hash: bytes = field(metadata={"store.type": "primary"})
    datetime: datetime


@dataclass
class FunctionCall:
    call_hash: bytes = field(metadata={"store.type": "primary"})
    function_hash: bytes
    kwargs: bytes
    datetime: datetime


@dataclass
class FunctionCallResult:
    call_hash: bytes
    result_type: bytes
    result: bytes
    computation_time: float
    time: datetime
    overrideable: bool = False



Chatter = Union[FunctionDefinition, FunctionCall, FunctionCallResult]


@dataclass(slots=True)
class FunctionCallProxy:
    call_hash: bytes


def hash_function_source(f):
    source = inspect.getsource(f)
    s = sha3_256(source.encode("utf-8"))
    hash = s.digest()
    return hash


class Vault:

    NO_RESULT = object()

    def __init__(self, storepath: Union[Store, str]):
        if isinstance(storepath, Store):
            self._store = storepath
        else:
            self._store = Store(storepath)
        self._store._register_dataclass(FunctionDefinition)
        self._store._register_dataclass(FunctionCall)
        self._store._register_dataclass(FunctionCallResult)
        self._store.create_db()
        self._function_store = {}
        self._store_function = False

    def consume(self, x: Chatter) -> None:
        if isinstance(x, FunctionDefinition):
            self._store.insert(x)
        elif isinstance(x, FunctionCall):
            self._store.insert(x)
        elif isinstance(x, FunctionCallResult):
            self._store.insert(x)
        else:
            raise ValueError(f'Can only handle values of type FunctionDefinition, FunctionCall or FunctionCallResult, not {type(x)}')

    def _get_store_def(self):
        with self._store:
            if not self._store_function:
                function_hash = b"vault.store"
                fdef = self.recall_definition_from_hash(function_hash)
                if fdef:
                    self._store_function = fdef
                else:
                    fdef = FunctionDefinition("vault.store", "vault internal hack", function_hash, datetime.now())
                    self._store.insert(fdef)
                    self._store_function = fdef
        return self._store_function

    def _calculate_store_variable_call(self, name):
        store_def = self._get_store_def()

        pickled_kwargs = pickle.dumps({"name": name}, protocol=5)
        call_hash = calculate_call_hash(store_def, pickled_kwargs)

        now = datetime.now()
        fcall = FunctionCall(call_hash=call_hash, function_hash=store_def.function_hash, kwargs=pickled_kwargs, datetime=now)
        return fcall

    def store_variable(self, name, value, final=False):
        with self._store:
            fcall = self._calculate_store_variable_call(name)
            print(f"Storing variable {name} with value {value} on callhash {fcall.call_hash}")
            if not self._store.get(FunctionCall, call_hash=fcall.call_hash):
                self._store.insert(fcall)
            self.save_result(fcall, value, 0.0, not final)

    def load_variable(self, name):
        fcall = self._calculate_store_variable_call(name)

        result = store.table(FunctionCallResult)
        correct_callhash = (store.col("call_hash") == fcall.call_hash)
        creation_time = store.col("time")

        results = self._store.select(result.where(correct_callhash).order_by(creation_time, "desc"), limit=1)
        if results:
            result = results[0]
            return self.load_result(result)
        else:
            return None

    def define(self, f: Callable) -> FunctionDefinition:
        source = inspect.getsource(f)
        function_hash = hash_function_source(f)
        fun_def = self._store.get(FunctionDefinition, function_hash=function_hash)
        if fun_def:
            fun_def = fun_def[0]
        else:
            name = f'{f.__module__}.{f.__name__}'
            fun_def = FunctionDefinition(name, source, function_hash, datetime.now())
            self._store.insert(fun_def)
        return fun_def

    def recall_definition_from_hash(self, function_hash: bytes) -> FunctionDefinition:
        return self._store.get_single(FunctionDefinition, function_hash=function_hash)

    def recall_definition_from_name(self, function_name: str) -> FunctionDefinition:
        return self._store.get(FunctionDefinition, name=function_name)

    def call(self, function_definition: FunctionDefinition, kwargs: Dict[str, Any]) -> FunctionCall:

        kwargs = proxy_function_calls(kwargs)
        pickled_kwargs = pickle.dumps(kwargs, protocol=5)

        call_hash = calculate_call_hash(function_definition, pickled_kwargs)

        fhash = function_definition.function_hash

        now = datetime.now()

        fcall = FunctionCall(call_hash=call_hash, function_hash=fhash, kwargs=pickled_kwargs, datetime=now)

        try:
            self._store.insert(fcall)
        except sqlite3.IntegrityError:
            pass
        return fcall

    def get_newest_unhandled_calls(self, limit: int=1) -> List[FunctionCall]:
        call = store.table(FunctionCall)
        result = store.table(FunctionCallResult)
        result_present = (store.col("result") == None)
        creation_time = store.col("time")

        calls = self._store.select(call.join(result, on="call_hash", how="left").where(result_present).order_by(creation_time, "desc"), limit=limit)
        return calls

    def get_result(self, function_call: Union[FunctionCall, FunctionCallProxy]):
        results = self._store.get(FunctionCallResult, call_hash=function_call.call_hash)
        results = [result for result in results if result.result_type != 'intent']
        if len(results) > 1:
            raise Exception("Got multiple results for the same function call, something seems wrong")
        elif len(results) == 1:
            return results[0]
        else:
            return Vault.NO_RESULT

    def save_result(self, function_call: FunctionCall, result: Any, time: float, overrideable: bool=False) -> FunctionCallResult:
        result_type = 'pickle'
        b = pickle.dumps(result, protocol=5)
        call_hash = function_call.call_hash
        fun_call_result = FunctionCallResult(call_hash, result_type, b, time, datetime.now(), overrideable)

        previous_results = self._store.get(FunctionCallResult, call_hash=call_hash)

        if len(previous_results) > 0:
            if previous_results[0].overrideable:
                self._store.insert(fun_call_result, update=True)
                return fun_call_result
            else:
                print("A result was already present, discarding the new one")
                return previous_results[0]
        else:
            self._store.insert(fun_call_result)
            return fun_call_result

    def load_result(self, result: FunctionCallResult) -> Any:
        if result.result_type == 'pickle':
            return pickle.loads(result.result)
        else:
            return result

    def lock_call(self, function_call: FunctionCall) -> FunctionCallResult:
        result_type = 'intent'
        b = b""
        call_hash = function_call.call_hash
        dt = 0
        fun_call_intent = FunctionCallResult(call_hash, result_type, b, dt, datetime.now(), True)
        self._store.insert(fun_call_intent)
        return fun_call_intent


def proxy_function_calls(value):
    if isinstance(value, list):
        return [proxy_function_calls(inner) for inner in value]
    elif isinstance(value, dict):
        return {key: proxy_function_calls(inner) for key, inner in value.items()}
    elif isinstance(value, FunctionCall):
        return FunctionCallProxy(value.hash)
    else:
        return value


def _create_kwargs(f, *args: Any, **kwargs: Any):
    params = inspect.signature(f).parameters
    kwargs.update({param: proxy_function_calls(arg) for param, arg in zip(params, args)})
    return kwargs



def calculate_call_hash(function_definition, pickled_kwargs: bytes):
    hasher = sha3_256()
    hasher.update(function_definition.function_hash)
    hasher.update(pickled_kwargs)
    hash = hasher.digest()
    return hash





import click
import logging

from .worker import Worker
from .server import Server

@click.group()
@click.version_option(version='0.1')
@click.option("--debuglevel", default="INFO")
def remoter(debuglevel):
    logging.basicConfig(format="%(asctime)s - %(name)s - %(process)s - %(levelname)s - %(message)s", level=debuglevel)

@remoter.command()
@click.argument('address', default="ipc://.remoter.ipc")
@click.option("--logfile", default=None)
def worker(address, logfile):
    if logfile:
        import logging
        root_logger = logging.getLogger()
        stream_handler = logging.FileHandler(logfile)
        root_logger.addHandler(stream_handler)
    worker = Worker(address)
    worker.start_async()

@remoter.command()
@click.argument('address', default="ipc://.remoter.ipc")
def server(address):
    server = Server(address)
    server.start_async()

if __name__ == "__main__":
    remoter()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['phantombuster', 'phantombuster.remoter']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'pandas>=2.0,<3.0',
 'polars>=1.0,<2.0',
 'pyarrow>=15,<16',
 'pysam>=0.22.0,<0.23.0',
 'regex>=2022.10.31,<2023.0.0',
 'scipy>=1.10.1,<2.0.0',
 'trio>=0.22.0,<0.23.0',
 'zmq>=0.0.0,<0.0.1']

entry_points = \
{'console_scripts': ['phantombuster = phantombuster.cli:phantombuster']}

setup_kwargs = {
    'name': 'phantombuster',
    'version': '0.16.1',
    'description': 'Bioinformatical tool to remove sequencing artifacts originating from single-nucleotide errors and index hopping from barcode-based experiments',
    'long_description': 'PhantomBuster is a bioinformatical tool that removes phantom barcode combinations that occur due to single-nucleotide sequencing errors and index hopping.\nIt is written for lineage-tracing experiments and CRISPR-screens, but can be used for any experimental setups in which only barcodes and no genetic DNA is measured.\n\n# Installation\n\nPhantomBuster is available via [pypi](https://pypi.org/project/phantombuster/) and can be installed with standard python tools like pip or pipx.\n\n    pipx install phantombuster\n\n# QuickStart\n\nPhantomBuster is a command line tool which can be run via the `phantombuster` command.\nIt consists of four main steps: (1) demultplexing, (2) error correction of random barcodes, (3) hopping removal and (4) thresholding.\nFor CRISPR-screens a separate script can calculate p-values for guides.\n\n## Demultiplexing\n\nPhantomBuster demultiplexes BAM or FASTQ files, extracts all specified barcodes while error correcting barcodes with known reference sequences.\nFor demultiplexing additional worker processes must be started.\n\n```\nphantombuster demultiplex [INPUTFILE] --outdir [DIR] --barcode-hierarchy-file [FILE] --regex-file [FILE] \nphantombuster worker --outdir [DIR]\n```\n\nINPUTFILE must be a csv file that lists all BAM and FASTQ files that are processed.\n\nExample `input_files.csv`:\n```\nfile\n101.bam\n```\n\nThe barcode hierarchy file is a csv file that lists all barcodes to be extracted.\nThe order of the barcodes creates a hierarchy, in which barcodes higher up the hierarchy are more general, while barcodes lower in the hierarchy are more specific.\nThe hierarchy is used in the second step error correction, in which two random barcode sequences are only compared, if all barcode sequences higher up the hierarchy are the same.\nExample `barcode_hierarchy.csv`\n\n```\nbarcode,type,referencefile,threshold,min_length,max_length\nsample,reference,sample_barcodes.csv,auto,-,-\nlib,reference,library_barcodes.csv,1,-,-\nlid,random,-,-,50,50\n```\n\nThe regex file is a csv file that specifies for each read region how to extract barcodes by a regular expression.\n\nExample `regexes.csv`:\n```\ntag,regex\nb2,"^[ACGTN]{3}(?P<sample>[ACGTN]{5})"\nquery,"(?P<lid>[ACGTN]{5,6}(?P<lib>ACGT|GTAC){s<=1}[ACGTN]+)"\n```\n\nThe outdir is a directory that contains all output and temporary files.\nThe same out directory must be passed to all stages of phantombuster.\n\n## Error Correction\n\nThe error correction step employs the UMI-tools error correction algorithm to error correct random barcode sequences.\nFor error correction additional worker processes must be started.\n\n```\nphantombuster error-correct --outdir [DIR] --barcode-hierarchy-file [FILE]\nphantombuster worker --outdir [DIR]\n```\n\nThe out dir and barcode hierarchy file must be the same as in the demultiplexing step.\n\n## Index Hopping Removal\n\nThe index hopping removal step removes barcode combinations that likely arised due to index hopping.\nFor index hopping removal no worker processes need to be started.\n\n```\nphantombuster hopping-removal --outdir [DIR] [HOPPING_BARCODES]\n```\n\nThe out dir must be the same as in the previous steps.\nThe barcodes to test must correspond to one or a combination of barcodes of the barcode hierarchy (`sample`).\nCombinations can be given seperated by commas (`sample,lid`).\nMultiple barcodes or combinations can be given and are then processed one after another (`sample,lid lib`)\nIt is recommended to test the combination of barcodes on the i5 index and then the combination of barcodes on the i7 index.\n\n## Thresholding\n\nThresholding removes all barcode combinations with a read count below a user defined threshold.\nSeperate Thresholds can be chosen for different values of barcodes, for example for each sample.\nFor tresholding no worker processes need to be started.\n\n```\nphantombuster threshold --outdir [DIR] --threshold-file [FILE]\n```\n\nExample `thresholds.csv`:\n```\nsample,threshold\nsample1,10\n```\n',
    'author': 'Simon Haendeler',
    'author_email': 'simon.emanuel.haendeler@univie.ac.at',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.13',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



import json
import logging
import zmq
from zmq.utils.strtypes import bytes, unicode, cast_bytes

LOG_RECORDS_ATTRIBUTES = ["args", "asctime", "created", "exc_info", "filename", "funcName", "levelname", "levelno", "lineno", "message", "module", "msecs", "msg", "name", "pathname", "process", "processName", "relativeCreated", "stack_info", "thread", "threadName"]

class PushHandler(logging.Handler):
    """A basic logging handler that emits log messages through a Push socket.
    Takes a PUB socket already bound to interfaces or an interface to bind to.
    Example::
        sock = context.socket(zmq.PUSH)
        sock.bind('inproc://log')
        handler = PushHandler(sock)
    Or::
        handler = PushHandler('inproc://loc')
    These are equivalent.
    """

    socket = None

    def __init__(self, interface_or_socket, context=None):
        logging.Handler.__init__(self)
        self.formatters = {
            logging.DEBUG: logging.Formatter(
                "%(levelname)s %(filename)s:%(lineno)d - %(message)s\n"
            ),
            logging.INFO: logging.Formatter("%(message)s\n"),
            logging.WARN: logging.Formatter(
                "%(levelname)s %(filename)s:%(lineno)d - %(message)s\n"
            ),
            logging.ERROR: logging.Formatter(
                "%(levelname)s %(filename)s:%(lineno)d - %(message)s - %(exc_info)s\n"
            ),
            logging.CRITICAL: logging.Formatter(
                "%(levelname)s %(filename)s:%(lineno)d - %(message)s\n"
            ),
        }
        if isinstance(interface_or_socket, zmq.Socket):
            self.socket = interface_or_socket
            self.ctx = self.socket.context
        else:
            self.ctx = context or zmq.Context()
            self.socket = self.ctx.socket(zmq.PUSH)
            self.socket.connect(interface_or_socket)

    def setFormatter(self, fmt, level=logging.NOTSET):
        """Set the Formatter for this handler.
        If no level is provided, the same format is used for all levels. This
        will overwrite all selective formatters set in the object constructor.
        """
        if level == logging.NOTSET:
            for fmt_level in self.formatters.keys():
                self.formatters[fmt_level] = fmt
        else:
            self.formatters[level] = fmt

    def format(self, record):
        """Format a record."""
        return self.formatters[record.levelno].format(record)

    def emit(self, record):
        """Emit a log message on my socket."""
        ei = record.exc_info
        if ei:
            # just to get traceback text into record.exc_text ...
            dummy = self.format(record)
        # See issue #14436: If msg or args are objects, they may not be
        # available on the receiving end. So we convert the msg % args
        # to a string, save it as msg and zap the args.
        d = dict(record.__dict__)
        d['msg'] = record.getMessage()
        d['args'] = None
        d['exc_info'] = None
        # Issue #25685: delete 'message' if present: redundant with 'msg'
        d.pop('message', None)

        self.socket.send_json(d)
        #self.socket.send(bmsg)






from __future__ import annotations

import inspect
from hashlib import sha256
import json
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, NoReturn, Optional, Set, Union, cast
import logging
from datetime import datetime
from types import TracebackType

from .messages import ResultMsg, TaskMsg, FSaveMsg, RemoterException
from .serialization import serialize_kwargs

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .store import Store
    from .persisters import Persister

# ==== Task ====


def depends(*fs):
    def wrapper(f):
        if not hasattr(f, "_remoter_depends"):
            f._remoter_depends = []
        for dep in fs:
            f._remoter_depends.append(dep)
        return f
    return wrapper


def hash_function(f):
    """Hash the source of the function f"""
    source = inspect.getsource(f)
    s = sha256(source.encode("utf-8"))
    if hasattr(f, "_remoter_depends"):
        for x in f._remoter_depends:
            s.update(hash_function(x).encode("utf8"))
    return s.hexdigest()


@dataclass
class Task:
    persister: Optional[Persister]
    store: Store
    datetime: datetime
    function_module: str
    function_name: str
    function_hash: str
    kwargs: Dict[str, Any]
    result: Optional[Any] = None
    task_hash: Optional[str] = None
    done: bool = False
    ignore_hash = False
    loaded = False
    exception: Optional[tuple[BaseException, TracebackType]] = None
    failed: bool = False
    time: Optional[float] = None
    code: Optional[str] = None

    def calc_hash(self):
        if self.task_hash is None:
            s = sha256()
            s.update(self.function_module.encode("utf8"))
            s.update(self.function_name.encode("utf8"))
            s.update(serialize_kwargs(self.kwargs).encode("utf8"))
            if not self.ignore_hash:
                s.update(self.function_hash.encode("utf8"))
            self.task_hash = s.hexdigest()
        return self.task_hash

    def to_result_message(self, worker_name: str) -> ResultMsg:
        assert self.task_hash
        return ResultMsg(
            task_hash=self.task_hash,
            datetime=self.datetime,
            function_module=self.function_module,
            function_name=self.function_name,
            kwargs=self.kwargs,
            result=None,
            worker_name=worker_name,
        )

    def to_task_message(self, name: str, request=True) -> TaskMsg:
        self.calc_hash()
        assert self.task_hash
        return TaskMsg(
            name=name,
            persister=self.persister.path if self.persister else None,
            task_hash=self.task_hash,
            datetime=self.datetime,
            function_module=self.function_module,
            function_name=self.function_name,
            function_hash=self.function_hash,
            kwargs=self.kwargs,
            failed=self.failed,
            exception=None,
            request=request,
            time=self.time
        )

    @classmethod
    def from_task_message(cls, message, persister, store) -> 'Task':
        task = Task(
            persister,
            store,
            message.datetime,
            message.function_module,
            message.function_name,
            message.function_hash,
            message.kwargs,
            result=None,
            task_hash=message.task_hash,
            done=False,
            time=message.time
        )
        return task

    def to_fsave_message(self):
        return FSaveMsg(
            datetime=self.datetime,
            function_module=self.function_module,
            function_name=self.function_name,
            function_hash=self.function_hash,
            function_code=self.code
        )

    def load(self):
        if self.loaded:
            return self.result
        self.calc_hash()
        logging.debug(f"Trying to load task {self}")
        try:
            r = self.store.load(self)
            self.result = r
            self.loaded = True
        except RemoterException as e:
            logging.debug("Loading of task failed %s", e)
            return False
        return self.result

    def load_metadata(self):
        try:
            msg = self.store._load_metadata(self)
        except RemoterException:
            return False
        return msg

    def is_loadable(self):
        try:
            self.store._load_metadata(self)
        except RemoterException:
            return False
        return True

    async def save(self):
        r = await self.store.save(self)

    async def wait(self):
        if not self.loaded:
            self.load()
        return self.result

    def _save_force(self):
        self.store._save_force(self)

    def __str__(self):
        kwargs = str(self.kwargs)
        if len(kwargs) > 20:
            kwargs = kwargs[:10] + "..." + kwargs[-10:]
        return f"Task({self.task_hash}, {self.function_module}, {self.function_name}, {self.function_hash}, {kwargs}, result=..., done={self.done})"

    def __repr__(self):
        kwargs = str(self.kwargs)
        if len(kwargs) > 20:
            kwargs = kwargs[:10] + "..." + kwargs[-10:]
        return f"Task({self.task_hash}, done={self.done})"

    def __eq__(self, o):
        if not isinstance(o, Task):
            return False
        return self.calc_hash() == o.calc_hash()

    def __hash__(self):
        v = int(self.calc_hash(), 16)
        return v



from typing import Any, Callable, Dict, List, NoReturn, Optional, Set, Union, cast
import uuid
import logging
import importlib
import inspect
import sys
import time

from . import globaladdress
from .store import Store, Remote
from .task import Task
from .async_ import async_run, async_sleep
from .messages import FreeMsg, Message, OKMsg, TaskMsg, QuitMsg, LockMsg, WaitMsg, ResultMsg, StoreMsg, FailMsg, ExceptionMsg, assert_never, parse_json_to_message, message_to_json, JoinMsg, FSaveMsg
from .persisters import persister_factory, InbasePersister
from .socket_ import ReqSocket

class Worker:

    instances: List["Worker"] = []

    def __init__(self, address: str = "ipc://.remoter.ipc", name: Optional[str] = None) -> None:
        if name:
            self._name = name
        else:
            self._name = uuid.uuid4().hex
        self._modules: dict[str, Any] = {}
        self._address: str = address
        self._socket: Optional[ReqSocket] = None
        self._running: bool = False
        self._result: Optional[Any] = None
        self._result_computed: bool = False
        self._task: Optional[Task]
        self._store = Store(":memory:", [Remote(self, name=self._name)])
        self._wait = None
        self._first_time_free = False

        globaladdress.global_address = address

        Worker.instances.append(self)
        self.logger = logging.getLogger(f"remoter.client.{self._name}")

    def __del__(self) -> None:
        Worker.instances.remove(self)

    def start_async(self):
        async_run(self._start_async)

    async def _start_async(self) -> None:
        self._socket = sock = ReqSocket()
        self.logger.info("Connecting to %s", self._address)
        sock.connect(self._address)
        self.logger.info("Connected successfully to server at %s", self._address)

        self._running = True
        self._first_time_free = True


        try:
            while self._running:
                freemsg = FreeMsg(self._name)
                if self._first_time_free:
                    self.logger.debug("Sending a message indicating I'm free")
                    self._first_time_free = False
                await self.send_async(freemsg)
                try:
                    msg = await self.recv_async()
                except TimeoutError as e:
                    self._running = False
                    self.logger.error("Did not receive an answer to my FreeMsg in time. I'm quitting without notifying the server")
                    continue
                await self.handle_message(msg)
        finally:
            self._socket.close()
        self.logger.info("Client shut down done")


    async def handle_message(self, message: Message) -> None:
        if isinstance(message, FreeMsg):
            raise Exception("Worker got a Free Message")
        elif isinstance(message, OKMsg):
            raise Exception("Worker got an OK Message")
        elif isinstance(message, TaskMsg):
            await self._handle_task(message)
            self._first_time_free = True
        elif isinstance(message, QuitMsg):
            self.logger.info("Worker was told to quit")
            self._running = False
        elif isinstance(message, LockMsg):
            raise Exception("Worker got an Lock Message")
        elif isinstance(message, WaitMsg):
            await self._handle_wait(message)
        elif isinstance(message, ResultMsg):
            raise Exception("Worker got an Result Message")
        elif isinstance(message, StoreMsg):
            raise Exception("Worker got an Store Message")
        elif isinstance(message, FailMsg):
            raise Exception("Worker got an Fail Message")
        elif isinstance(message, ExceptionMsg):
            raise Exception("Worker got an Exception Message")
        elif isinstance(message, JoinMsg):
            raise Exception("Worker got an JoinMessage")
        elif isinstance(message, FSaveMsg):
            raise Exception("Worker got an JoinMessage")
        else:
            assert_never(message)

    async def load_kwargs(self, kwargs):
        def find_tasks(value):
            if isinstance(value, Task):
                return [value]
            elif isinstance(value, list):
                l = []
                for v in value:
                    for task in find_tasks(v):
                        l.append(task)
                return l
            elif isinstance(value, dict):
                l = []
                for v in value.values():
                    for task in find_tasks(v):
                        l.append(task)
                return l
            else:
                return []

        tasks = find_tasks(kwargs)

        for task in tasks:
            await self.send_async(task.to_task_message(name=self._name, request=False))
            try:
                storemsg = await self.recv_async()
                self._store._income(storemsg)
                task.load()
            except TimeoutError as e:
                self._running = False
                logging.error("Did not receive an answer to my TaskMsg, is the server done? Quitting.")
                break

    async def _handle_task(self, message):
        await self.load_kwargs(message.kwargs)
        if not self._running:
            return
        self.logger.info("Worker is executing task %s with hash %s", message.function_name, message.task_hash)
        persister = persister_factory(message.persister)
        self.logger.debug("Using persister %s for task %s", persister, message)
        task = Task(
            persister,
            self._store,
            message.datetime,
            message.function_module,
            message.function_name,
            message.function_hash,
            message.kwargs,
            result=None,
            task_hash=message.task_hash,
            done=False,
            time=None
        )
        task = await self.execute(task)
        await task.save()
        await self.send_async(task.to_result_message(self._name))
        try:
            await self.recv_async()
        except TimeoutError as e:
            logging.error("Did not receive an answer to my ResultMsg, is the server done? Continuing anyway.")


    async def _handle_wait(self, message):
        return await async_sleep(message.time)

    async def handle_exception_async(self, e) -> None:
        #await self.send_async(ExceptionMsg(self._name, json.dumps(tblib.Traceback(e[2]).to_dict())))
        await self.send_async(ExceptionMsg(self._name, str(e)))
        try:
            await self.recv_async()  # TODO check result was okay
        except TimeoutError as e:
            logging.error("Did not receive an answer to my ExceptionMsg, is the server done? Continuing anyway.")

    def _load_value(self, value):
        if isinstance(value, Task):
            return value.load()
        elif isinstance(value, list):
            return [self._load_value(v) for v in value]
        elif isinstance(value, dict):
            return {key: self._load_value(v) for key, v in value.items()}
        else:
            return value


    async def execute(self, task: Task) -> Task:
        if task.function_module not in self._modules:
            self._modules[task.function_module] = importlib.import_module(task.function_module)
        f = getattr(self._modules[task.function_module], task.function_name)
        self.logger.debug("Executing task")
        start_time = time.time()
        end_time = None
        kwargs = self._load_value(task.kwargs)
        try:
            start_time = time.time()
            result_or_awaitable = f(**kwargs)
            end_time = time.time()
            if inspect.isawaitable(result_or_awaitable):
                start_time = time.time()
                result = await result_or_awaitable
                end_time = time.time()
            else:
                result = result_or_awaitable
            task.done = True
            task.result = result
            task.task_hash = task.calc_hash()
            task.time = end_time - start_time
        except Exception as e:
            if end_time is None:
                end_time = time.time()
            self.logger.exception("Client encountered an exception when executing a task.")
            await self.handle_exception_async(e)
            task.failed = True
            task.done = True
            task.time = end_time - start_time
            err = sys.exc_info()
            task.exception = (err[1], err[2]) if err[1] and err[2] else None
            p = InbasePersister()
            task.persister = p
        return task

    async def recv_async(self) -> Message:
        if self._socket is not None:
            self.logger.debug("Waiting for an answer")
            msg_json = await self._socket.async_recv_json()
            msg = parse_json_to_message(msg_json, self._store)
            return msg
        else:
            raise Exception("Worker has to connect before being able to receive messages")

    async def send_async(self, msg: Message) -> None:
        if self._socket:
            self.logger.debug("Sending msg %s", msg)
            j = message_to_json(msg)
            await self._socket.async_send_json(j)
        else:
            raise Exception("Worker has to connect before being able to send messages")




import logging
import os
import os.path
import sys
import multiprocessing as mp
import glob

import ntpath
import pandas as pd
#from phantombuster import old
from dataclasses import dataclass
import json

import phantombuster.remoter
from phantombuster import porcelain, plumbing, stores
from phantombuster.stores import deduplicator_to_pyarrow_table
from phantombuster.remoter import Scheduler
from phantombuster.io_ import PathsAndFiles, write_parquet, write_csv
from phantombuster.project import Project
from phantombuster.config_files import read_barcode_hierarchy_file, read_input_files_file, read_regex_file
import phantombuster.statistics as statistics
import click
from typing import Optional, List
import pyarrow.parquet
import pyarrow.csv
import polars as pl

from pathlib import Path


def demultiplex(input_files_file, regex_file, barcode_hierarchy_file, project, debug=False, show_qc=False):
    project.create()

    input_groups = read_input_files_file(input_files_file)
    barcodes = read_barcode_hierarchy_file(barcode_hierarchy_file)
    barcodes_names = [bc['name'] for bc in barcodes]
    regexes = read_regex_file(regex_file)

    for bc_name in barcodes_names:
        names_with_numbers = []
        if bc_name.rstrip("0123456789") != bc_name:
            names_with_numbers.append(bc_name)
    if len(names_with_numbers) > 0:
        logging.error(f'Barcode names missconfigured. barcode names cannot have numbers at the end. Problematic barcode names: {names_with_numbers}.')
        raise Exception(f'Barcodes missconfigured: {names_with_numbers}')
           
    missconfigured_groups = []
    for input_group in input_groups:
        rs = regexes.get_regexes_for_group(input_group.group)
        r_tags = [name for r in rs.values() for name in r.groupindex.keys()]
        r_tags  = list(set([tag.rstrip('0123456789') for tag in r_tags]))

        missing_tags = [bc for bc in barcodes_names if bc not in r_tags]
        superflous_tags = [tag for tag in r_tags if tag not in barcodes_names]
        if len(missing_tags) > 0 or len(superflous_tags) > 0:
            missconfigured_groups.append((input_group, missing_tags, superflous_tags))

    if len(missconfigured_groups) > 0:
        for g, missing, superflous in missconfigured_groups:
            logging.error(f'Group {g.group} with example file {g.files[0].path} missconfigured. Missing tags: {missing}. Unused tags: {superflous}.')
        raise Exception('Groups missconfigured')


    phantombuster.remoter.load(stores.load, dir=str(project.tmp_dir))(plumbing.deduplicate_section)
    phantombuster.remoter.save(stores.save, dir=str(project.tmp_dir))(plumbing.deduplicate_section)

    phantombuster.remoter.load(stores.load, dir=str(project.tmp_dir))(plumbing.combine)
    phantombuster.remoter.save(stores.save, dir=str(project.tmp_dir))(plumbing.combine)

    scheduler = project.get_scheduler()

    with scheduler:
        maintable, (reads, remaining_reads, counters_success, counters_fail) = porcelain.demultiplex(input_files_file, project.demultiplex_output_path, project.demultiplex_stats_path, regex_file, barcode_hierarchy_file, debug, show_qc, scheduler)
    m = max([len(str(count)) for count in counters_fail.values()] + [len(str(reads))])

    logging.info(f"{str(reads).rjust(m)} Reads processed")
    logging.info(f"{str(remaining_reads).rjust(m)} Reads Remain ({100*remaining_reads/reads}%)")
    for reason, count in counters_fail.items():
        logging.info(f"{str(count).rjust(m)} Reads were unfit due to {reason}")

    r = (maintable, (reads, remaining_reads, counters_success, counters_fail))
    print(maintable)
    return r


def error_correct(project, error_threshold, barcode_hierarchy_file, remove_ambigious):
    project.create()

    phantombuster.remoter.load(stores.load, dir=str(project.tmp_dir))(porcelain.error_correct_partition)
    phantombuster.remoter.save(stores.save, dir=str(project.tmp_dir))(porcelain.error_correct_partition)

    barcode_hierarchy = read_barcode_hierarchy_file(barcode_hierarchy_file)

    table_file = project.demultiplex_output_path
    out_file = project.error_correct_output_path

    error_corrected = porcelain.error_correct(table_file, out_file, error_threshold, barcode_hierarchy, project, remove_ambigious)

    print(f"LIDs after correction: {error_corrected['lids_corrected']}")
    print(f"Reads after correction: {error_corrected['reads_corrected']}")

    with open(project.error_correct_stats_path, mode="w") as f:
        json.dump(error_corrected, f)

    logging.info('Error correction done')


def hopping_removal(project, hopping_barcodes, alpha_threshold):
    project.create()

    table_file = project.error_correct_output_path

    r, stats = porcelain.hopping_removal(table_file, alpha_threshold, hopping_barcodes)

    print(f"LIDs before hopping removal: {stats['lids_before']}")
    print(f"LIDs after hopping removal: {stats['lids_after']}")
    print(f"Reads before hopping removal: {stats['reads_before']}")
    print(f"Reads after hopping removal: {stats['reads_after']}")

    write_parquet(r, project.hopping_removal_output_path)
    with open(project.hopping_removal_stats_path, mode='w') as f:
        json.dump(stats, f)


def threshold(project, threshold_file):
    project.create()

    # read in threshold file
    thresholds = porcelain.read_file(threshold_file)

    table, stats = porcelain.threshold(project, threshold_file)

    write_parquet(table, project.threshold_output_path)
    with open(project.threshold_stats_path, mode='w') as f:
        json.dump(stats, f)


def pvalue(project, read_file, category_file, column='reads', group_by=['grna'], N=1_000_000, max_n=950):
    project.create()
    categories = pl.DataFrame(porcelain.read_file(category_file))
    table = pl.read_parquet(read_file)

    join_columns = [col for col in categories.columns if col != 'category']

    table = table.join(categories, on=join_columns, validate='m:1')

    results = statistics.calculate_pvalues(table, column=column, group_by=group_by, N=N, max_n=max_n)
    results = results.join(categories, on=join_columns, validate='1:1')
    write_csv(results, project.pvalue_output_path)
